;  $Id: LDRBEXT.ASM 1.6 2003/03/20 17:19:26 thomd Exp $
;
;  *** DO NOT REMOVE THE FOLLOWING NOTICE ***
;
;  Copyright (c) 1991-2003 Datalight, Inc.
;  All Rights Reserved
;
;  Datalight, Inc. is a Washington corporation located at:
;
;  21520 30th Dr SE, Suite 110,          Tel: +1.425.951.8086
;  Bothell, WA  98021                    Fax: +1.425.951.8094
;  USA                                   Http://www.datalight.com
;
;  This software, including without limitation all source code and
;  documentation, is the confidential, trade secret property of
;  Datalight, Inc., and is protected under the copyright laws of
;  the United States and other jurisdictions.  Portions of the
;  software may also be subject to one or more the following US
;  patents: US#5860082, US#6260156.
;
;  In addition to civil penalties for infringement of copyright
;  under applicable U.S. law, 17 U.S.C. 1204 provides criminal
;  penalties for violation of (a) the restrictions on circumvention
;  of copyright protection systems found in 17 U.S.C. 1201 and
;  (b) the protections for the integrity of copyright management
;  information found in 17 U.S.C. 1202.
;
;  U.S. Government Restricted Rights.  Use, duplication,
;  reproduction, or transfer of this commercial product and
;  accompanying documentation is restricted in accordance with
;  FAR 12.212 and DFARS 227.7202 and by a License Agreement.
;
;  IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
;  PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
;  AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
;  INC. AND THE LICENSEE ("BINDING AGREEMENT").  YOUR RIGHT, IF ANY,
;  TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
;  TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
;  USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
;  TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT.  CONTACT
;  DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
;  THIS SOFTWARE IN ERROR.
;
;--------------------------------------------------------------------
; $DOCMODULE:LDRBEXT.ASM
;
; Copyright (c) 1991-2002 Datalight, Inc.
; All Rights Reserved.
; Datalight Confidential and Proprietary
;
; A BIOS Extension that loads ROM-DOS.SYS from a Datalight ROM Disk.
;
; Steps:
; 1. The BIOS runs this BIOS extension (this module), which traps
;    Int19h. Displays a 'B' at this point.
; 2. The BIOS passes control to Int19h, "Int19Boot" routine in this
;    module. Displays a '-' at this point, then copies itself to the
;    top of RAM.
; 3. The Loader() function finds ROM-DOS.SYS on the paged ROM disk
;    and loads it at 70:0.
; 4. The Loader() routine then jumps to 70:0 (ROM-DOS) which then
;    boots the system just as if ROM-DOS was loaded from a floppy.
;    Displays numbers '0'-'9' at this point.
;
; $DOCHISTORY
; 04/20/2000 dg  Added ability to have the Bios extension in page 0
;                of a paged disk by copying all of the Bios extension
;                to RAM (so pages can be changed).
; 07/23/1998 dg  Updated to Solid Code
; 11/20/1991 dg  Created by Drew Gislason
;--------------------------------------------------------------------
include macros.inc

; Main Memory Locations
INT19H		equ	19H * 4		; boot vector
STORAGE		equ	21H * 4		; store old Int 19H vector
STACKSIZE	equ	1024		; size of stack in bytes

; enable BIOS putc
BCHECK		EQU	1

;--------------------------------------------------------------------
;   SEGMENT ORDER (for LINKER)
;--------------------------------------------------------------------
cBegCode
cEndCode
DGROUP GROUP _BEGDATA, _DATA, _CONST, _MSG, _BSS
_BEGDATA segment para public 'DATA'
_BEGDATA ends
_DATA	segment byte public 'DATA'
_DATA	ends
_CONST segment byte public 'CONST'	; CONST & MSG for Microsoft C
_CONST ends				; compatibility
_MSG segment byte public 'MSG'
_MSG ends
_BSS	segment byte public 'BSS'
_BSS	ends
_ENDBSS	segment para public 'BSS'
	EndBSS label word
_ENDBSS ends


;--------------------------------------------------------------------
;    DATA GROUP
;--------------------------------------------------------------------
cBegData
	BegData	label	word
	NullPtr		dd	0
	SomeText	db "Dgroup",0
cEndData

cBegUdata
	BegBSS	label	word

	stackBottom	db STACKSIZE dup (?)	; stack
	stackTop	label word
cEndUdata

;--------------------------------------------------------------------
; $DOCPROC:BiosExtension
;
; The BIOS extension part of the paged loader. The routine Loader()
; does the REAL work. This BIOS extension sets up Int 19H to point
; to our code so Loader() can run.
;--------------------------------------------------------------------
cBegCode
	extrn _Loader:near 		; C Routine to Load ROM-DOS

	public BiosExtension
BiosExtension proc far

	; The following 6 bytes MUST not be changed.  They enable the
	; BIOS to see this BIOS extension.  The checksum byte will be
	; modified during locate time (using LOC.EXE), after the
	; program has been linked.
	dw	0AA55h	; BIOS extension id
	db	4	; size of extension in 512 byte chunks
	jmp	short _astart
	db	0FFh	; checksum will be placed here by LOC.EXE

	; This ID is here simply to have some recognizable bytes in
	; the BIOS extension, so you can tell it's the paged loader
	; from DUMP or DEBUG.
	PUBLIC	_astart
_astart:
	jmp	short after_id
	db	"ROM-DOS Paged Loader",0

	; This BIOS extension code simply points Int 19H to
	; "Int19Boot" which runs the Loader() routine.
after_id:
	push	ds
	push	ax
	push	bx
	push	cx
	push	dx

	; point ds to vector table
	xor	ax, ax
	mov	ds, ax

	; save old Int 19H handler
	mov	ax, word ptr ds:[INT19H]
	mov	dx, word ptr ds:[INT19H+2]
	mov	word ptr ds:[STORAGE], ax
	mov	word ptr ds:[STORAGE+2], dx

	; set our new Int 19H handler
	mov	word ptr ds:[INT19H], offset Int19Boot
	mov	word ptr ds:[INT19H+2], cs

	; let outside world know the BIOS extension got control
IFDEF	BCHECK
	mov	al,'B'
	mov	ah,0Eh
	mov	bx, 7
	int	10h
ENDIF

	pop	dx
	pop	cx
	pop	bx
	pop	ax
	pop	ds
	ret
BiosExtension endp


;--------------------------------------------------------------------
; $DOCPROC:Int19Boot
;
; This routine gets control when the BIOS generates an Int19h. This
; creates a stack, copies data to RAM, and calls Loader().
;--------------------------------------------------------------------
Int19Boot:

	; let outside world know the Int 19H handler got control
IFDEF	BCHECK
	mov	al, '-'
	mov	ah, 0Eh
	mov	bh, 0
	int	10h
ENDIF

	; determine where to copy things in RAM
	int	12h	; returns kBytes RAM in ax
	mov	cl, 6
	shl	ax, cl	; convert to paragraphs (A000h = 640k)
	mov	dx, offset _TEXT:EndBSS
	mov	cl, 4
	shr	dx, cl	; convert size of Bios extension to paras
	sub	ax, dx	; top of memory
	mov	es, ax	; es now equals

	; copy Bios extension from to RAM
	mov	cx, cs
	mov	ds, cx
	xor	si, si
	xor	di, di
	mov	cx, offset _TEXT:BegBSS
	cld
	rep
	movsb

	; setup dgroup
	mov	dx, offset _TEXT:BegData	; get begining of DGROUP
	mov	cl, 4				; relative to code
	shr	dx, cl
	mov	ax, es		; es currently pointing to code
	add	ax, dx		; add in to start of dgroup
	mov	ds, ax		;

	; zero BSS (uninitialized C data)
	mov	di, offset _TEXT:BegBSS		; es still pointing to code
	mov	cx, offset _TEXT:EndBSS
	sub	cx, di
	xor	ax, ax
	cld
	rep
	stosb

	; setup stack in dgroup
	mov	ax, ds
	mov	dx, offset DGROUP:stackTop
	cli
	mov	ss, ax		; ss = dgroup
	mov	sp, dx
	sti

	; far jump to code down in RAM
	push	es		; es still pointing to code
	mov	ax, offset _TEXT:BackHere
	push	ax
	retf
    BackHere:

	; call Loader(). won't return if it found ROM-DOS
	call	_Loader

	; Loader() returned because ROM-DOS was not found. So jump
	; to original Int19h handler

	; restore old Int 19H handler
	xor	ax, ax
	mov	ds, ax
	mov	ax, word ptr ds:[STORAGE]
	mov	dx, word ptr ds:[STORAGE+2]
	mov	word ptr ds:[INT19H], ax
	mov	word ptr ds:[INT19H+2], dx

	; go do it!
	jmp	dword ptr ds:[INT19H]

cEndCode

	END 	BiosExtension

