//  $Id: MEMDISK.H 1.7 2003/03/20 16:58:56 thomd Exp $
//
//  *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//  Copyright (c) 1993-2003 Datalight, Inc.
//  All Rights Reserved
//
//  Datalight, Inc. is a Washington corporation located at:
//
//  21520 30th Dr SE, Suite 110,          Tel: +1.425.951.8086
//  Bothell, WA  98021                    Fax: +1.425.951.8094
//  USA                                   Http://www.datalight.com
//
//  This software, including without limitation all source code and
//  documentation, is the confidential, trade secret property of
//  Datalight, Inc., and is protected under the copyright laws of
//  the United States and other jurisdictions.  Portions of the
//  software may also be subject to one or more the following US
//  patents: US#5860082, US#6260156.
//
//  In addition to civil penalties for infringement of copyright
//  under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//  penalties for violation of (a) the restrictions on circumvention
//  of copyright protection systems found in 17 U.S.C. 1201 and
//  (b) the protections for the integrity of copyright management
//  information found in 17 U.S.C. 1202.
//
//  U.S. Government Restricted Rights.  Use, duplication,
//  reproduction, or transfer of this commercial product and
//  accompanying documentation is restricted in accordance with
//  FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//  IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//  PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//  AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//  INC. AND THE LICENSEE ("BINDING AGREEMENT").  YOUR RIGHT, IF ANY,
//  TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//  TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//  USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//  TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT.  CONTACT
//  DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//  THIS SOFTWARE IN ERROR.
//
/********************************************************************
	$DOCMODULE:MEMDISK.H

	Copyright (c) 1993-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary.

	This file must be included in each generic memory disk driver.
	While the name MEMDISK implies the interface only works with
	MEMORY disk drivers, it works fine as a base for ANY kind of disk
	driver.

	$DOCHISTORY
	07/23/1998 dg  Updated to Solid Code Conventions, added MEMUTIL.
	10/15/1993 dg  Created by Drew Gislason.
*********************************************************************/

/* new types */
typedef unsigned char uchar;
typedef unsigned long ulong;
typedef int bool;
#define FALSE 0
#define TRUE  !FALSE

/* assert is only defined in DEBUG builds of drivers */
#if DEBUG
void PASCAL AssertFail(int nFile, int nLine);
#define ASSERT(expr) if(!(expr)) AssertFail(__FILENUM__, __LINE__)
#define ABORT() AssertFail(__FILENUM__, __LINE__)
#else
#define ASSERT(expr)
#define ABORT()
#endif

/*
   BIOS PARAMETER BLOCK (BPB)

   Description: Every disk has this structure in the boot
   sector (sector 0).  The BPB describes the disk to DOS.
   DOS gets the BPB information both upon initialization,
   and any time the disk may have been removed and replaced
   with some other disk.

	Any decent DOS book describes the BPB.
*/
struct BPB /*()*/
{
	unsigned  BytesPerSector;	/* 00H calculated */
	uchar     SecPerCluster;	/* 02H calculated */
	unsigned  Reserved;			/* 03H */
	uchar     NumbFats;			/* 05H */
	unsigned  MaxDirs;			/* 06H calculated */
	unsigned  TotalSecs;			/* 08H calculated */
	uchar	    MediaDescrip;		/* 0AH */
	unsigned  SecPerFAT;			/* 0BH calculated */
	unsigned  SecPerTrak;		/* 0DH */
	unsigned  Sides;				/* 0FH */
   ulong     Hidden;				/* 11H */
   ulong     LNumSec;			/* 15H */
};
#define START_BPB		11

/*
   DIRECTORY ENTRY
   Description: Each item in the directory looks like this

   Directory attribute:
   xxxx xxxx xxxx xxx1    Read only
   xxxx xxxx xxxx xx1x    Hidden
   xxxx xxxx xxxx x1xx    System
   xxxx xxxx xxxx 1xxx    Volume label
   xxxx xxxx xxx1 xxxx    Subdirectory
   xxxx xxxx xx1x xxxx    Archive

   Reference: Adv MS-DOS, p 184  -or-  DOS Ref, p 248
*/
struct DIR
{
   char     name[8];         /* 00 file name                */
   char     ext[3];          /* 08 */
   uchar    attr;            /* 0B file attribute           */
   uchar    reserved[10];    /* 0C */
   unsigned time;            /* 16 file date                */
   unsigned date;            /* 18 file time                */
   unsigned start;           /* 1A starting cluster of file */
   ulong    size;            /* 1C file size in bytes       */
};


/* a far function */
typedef void (far *far_func_ptr)();

/*
	TSR Definition

	Description: Memory Disks can be configured to install
	as either a device driver or a TSR (which can then be
	unloaded if required).

	This structure allows a program to query a previously
	installed TSR Generic Disk.
*/
struct tsr_def /*()*/
{
	far_func_ptr tsr_uninit_addr;		/* a TSR "uninstall" routine */
	uchar tsr_start_drive;				/* starting drive of TSR */
	uchar tsr_num_drives;				/* # of drives in TSR */
	uchar tsr_hooked;						/* are the drives presently "active" */
	uchar tsr_id_len;						/* length of ID string */
	char *tsr_id_ptr;						/* the unique ID string itself */
};

/* mem_changed() return codes */
#define CHANGED_YES		-1
#define CHANGED_MAYBE	0
#define CHANGED_NO		1

/* some standard disk defaults */
#define DEF_SEC_SIZE		512		/* default sector size */
#define FAT_ID				0xF0		/* default FAT id byte */

/* categories for IOCTL calls */
#define BLOCK_IOCTL_READ  0xFF03
#define BLOCK_IOCTL_WRITE 0xFF0C

/* valid ROM-DISK identification */
#define ROMDISK_ID_OFF  3
#define ROMDISK_ID "DLRDISK"
#define SECTOR_ID   0xAA55		/* ID offset is determined by BPB */

/*
	Device Error Codes

	If the device must return an error, the following
	defines the possible error conditions:
*/
#define DEVERR_WRITEPROT	0x00	/* write protected */
#define DEVERR_BADUNIT		0x01	/* invalid unit # */
#define DEVERR_NOTREADY		0x02	/* disk not ready (drive door open) */
#define DEVERR_BADCMD		0x03
#define DEVERR_CRC			0x04
#define DEVERR_BADSTRUCT	0x05
#define DEVERR_SEEK			0x06
#define DEVERR_BADMEDIA		0x07
#define DEVERR_SECNOTFOUND	0x08	/* sector not found */
#define DEVERR_NOPAPER		0x09
#define DEVERR_WRITEFAULT	0x0A
#define DEVERR_READFAULT	0x0B
#define DEVERR_GENERALFAIL	0x0C


/* math functions from MEMUTIL.LIB */
ulong PASCAL _lshru(ulong l, unsigned u);
ulong PASCAL _lshlu(ulong l, unsigned u);
ulong PASCAL _lmulu(ulong l, unsigned u);
ulong PASCAL _ldivu(ulong ul, unsigned u);
ulong _ldiv(ulong lVal, ulong lDivideBy);
ulong _lmod(ulong lVal, ulong lModBy);

/* string functions from MEMUTIL.LIB */
void PASCAL memmove(void *to,void *from,unsigned len);
void PASCAL fmemcpy(void far *dst,void far *src,unsigned len);
void PASCAL memset(void *dst,uchar val,unsigned len);
void PASCAL fmemset(void far *dst,uchar val,unsigned len);
int PASCAL memcmp(char *str1,char *str2,unsigned len);
int PASCAL fmemcmp(char far *str1,char far *str2,unsigned len);
int PASCAL strlen(char *s);
int PASCAL fstrlen(char far * s);
char far * PASCAL MemSkipWhite(char far *lpStr);
char far * PASCAL MemSkipArg(char far *lpStr);

/* display functions from MEMUTIL.LIB */
int printf(char *fmt, ...);
int sprintf(char *szStr, char *fmt, ...);
void panic(char *fmt, ...);

/* address manipulation functions from MEMUTIL.LIB */
ulong PASCAL SegToLinear(unsigned seg);
void far * PASCAL AddPtr(unsigned seg, ulong lAddr);
ulong PASCAL LinearAddress(void *pPtr);
void far * PASCAL Linear2Ptr(ulong lAddress);
void far * PASCAL AddNormalize(void far *lpPtr, ulong lOffset);
unsigned PASCAL LOWORD(ulong lVal);
unsigned PASCAL HIWORD(ulong lVal);
ulong PASCAL MAKE_DWORD(unsigned hiWord, unsigned loWord);

/* ctype style functions from MEMUTIL.LIB */
char PASCAL MemToUpper(char c);
bool PASCAL MemIsDigit(char c);
bool PASCAL MemIsXDigit(char c);
int PASCAL MemXDigitVal(char c);

/* high-level disk functions from MEMUTIL.LIB */
bool FormatRAMDisk(int BytesPerSector, int DirEntries, unsigned NumSectors);
bool GetDiskBPB(struct BPB * pBPB);
bool IsValidDisk(void);
bool IsSameBPB( struct BPB * pBPB1, struct BPB * pBPB2 );
bool RamDiskUnitTest(char szCmdline[]);
bool RomDiskUnitTest(char szCmdline[]);

/* misc functions from MEMUTIL.LIB */
unsigned PASCAL WordCheckSum(void far *lpBuff, unsigned nLen);
unsigned PASCAL MinMax(char far *lpStr, unsigned nMin, unsigned nMax);
void OnInt19WipeVDisk(void);

/* extended memory routines */
bool CopyToInt15(ulong lDstAddr, void far *pSrc, unsigned nWords);
bool CopyFromInt15(void far *pDst, ulong lSrcAddr, unsigned nWords);
unsigned Int15MemSize(void);
bool IsFirstVDisk(void);

/* I/O routines */
void OutpByte(unsigned port, int val);
void OutpWord(unsigned port, int val);
int InpByte(unsigned port);
int InpWord(unsigned port);

/* public data from MEMUTIL.LIB */
extern char DiskIntro[11];
extern char aMemBuff[DEF_SEC_SIZE];

/* prototypes for MEMPAGED target */
extern char szPagedDiskIDString[];
extern long lPagedDiskWindowSize;
extern uchar far *lpPagedDiskWindow;
bool init_page(void);
bool set_page(unsigned nPage);
void uninit_page(void);

/* Generic Memory Disk Interface: global data */
extern uchar memerr;
extern uchar memunit;
extern unsigned memtop;

/* Generic Memory Disk Interface: user supplied functions */
int  meminit(struct BPB bpb[], char far *cmdline, unsigned *endseg, int drv);
void memuninit(void);
int  memchanged(struct BPB *bpb);
bool memread(long offset, unsigned len, char far * buffer);
bool memwrite(long offset, unsigned len, char far * buffer);
bool memioctl(unsigned category, char far * buffer);
