//  $Id: MEMFIXED.C 1.6 2003/03/20 16:55:22 thomd Exp $
//
//  *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//  Copyright (c) 1995-2003 Datalight, Inc.
//  All Rights Reserved
//
//  Datalight, Inc. is a Washington corporation located at:
//
//  21520 30th Dr SE, Suite 110,          Tel: +1.425.951.8086
//  Bothell, WA  98021                    Fax: +1.425.951.8094
//  USA                                   Http://www.datalight.com
//
//  This software, including without limitation all source code and
//  documentation, is the confidential, trade secret property of
//  Datalight, Inc., and is protected under the copyright laws of
//  the United States and other jurisdictions.  Portions of the
//  software may also be subject to one or more the following US
//  patents: US#5860082, US#6260156.
//
//  In addition to civil penalties for infringement of copyright
//  under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//  penalties for violation of (a) the restrictions on circumvention
//  of copyright protection systems found in 17 U.S.C. 1201 and
//  (b) the protections for the integrity of copyright management
//  information found in 17 U.S.C. 1202.
//
//  U.S. Government Restricted Rights.  Use, duplication,
//  reproduction, or transfer of this commercial product and
//  accompanying documentation is restricted in accordance with
//  FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//  IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//  PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//  AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//  INC. AND THE LICENSEE ("BINDING AGREEMENT").  YOUR RIGHT, IF ANY,
//  TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//  TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//  USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//  TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT.  CONTACT
//  DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//  THIS SOFTWARE IN ERROR.
//
/********************************************************************
	$DOCMODULE:MEMFIXED.C

	Copyright (c) 1995-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary.

	Fixed location conventional memory RAM disk.

	Syntax:
	DEVICE=MEMFIXED.SYS [DiskSize [SectorSize [NumEntries]]] [/F#] [/S]

	/F#			Fixed location for RAM disk (in segment form)
	/S				Check for static disk. Preserved across reboots

	DiskSize		Size of disk in Kilobytes. Valid range is 4 - 512.
	SectorSize	Size of each disk sector. Valid values are 128, 256 and 512.
	NumEntries	Number of root directory entries.

	This driver creates a fixed location RAM disk below 1 Megabyte.

	The /F# switch tells the driver where to locate the RAM disk. The
	# is a segment in hex form. Example1: /F2000 indicates to make a
	fixed RAM disk at 2000:0. The memory is be allocated from the DOS
	memory pool. Example2: /FD000 indicates to use RAM at D000:0.

	The disk may also be static (/S) across reboots. The static option
	only functions if the RAM is battery backed, or preserved in some
	other fashion across reboots.

	$DOCHISTORY
	07/23/1998 dg  Updated to Solid Code standards
	03/03/1995 dg  Created by Drew Gislason
**********************************************************/
#define __FILENUM__	201
#include "dosver.h"
#include "compiler.h"
#include "memdisk.h"

//#include "fatworld.h"
bool PASCAL dlIsFat32World( void );

/* some defaults */
#define DEF_DIRS  64			/* default # of root directory entries */
#define DEF_SIZEK 64			/* default size of RAM disk in KBytes */
#define VDISK_ID 0xF8		/* media ID, must be F8 for SCANDISK */

/* cmdline options and arguments */
bool	bStatic;								/* /S option  */

/* disk parameters */
unsigned nTotalSecs;
unsigned nDiskSizeK = DEF_SIZEK;
int	nBytesPerSector = DEF_SEC_SIZE;
int	nRootEntries = DEF_DIRS;

/* disk location */
unsigned nStartSeg = 0xD000;		/* if conventional RAM disk */

#define DEF_DIR_ENTRIES 	DEF_DIRS
struct BPB std_bpb =
{
	DEF_SEC_SIZE,1,1,2,DEF_DIR_ENTRIES,DEF_SIZEK*2,FAT_ID,6,9,2,0
};
char std_volume[] = "VDISK V6.22";
unsigned long std_serial = 0x12345678L;
char std_fattype[] = "FAT12    ";

/* messages */
char szSignOn[]    = "\nMEMFIXED %s %s\n%s\n\n";
char szErrFormat[] = "Format failed";
char szErrAlloc[]  = "Failed to find memory";
char szErrBadOpt[] = "Invalid option \"/%s\"!\n";
char szInstalled[] = "Installed %dKB %sRAM disk as drive %c: at %p:0\n";
char szStatic[]    = " (static) ";

/* default cmdline for built-in drivers */
char szBuiltinCmdline[] = "32 /FD000";

/* prototypes */
bool ParseCmdline(char far *lpCmdline);

bool fIsFat32World;

void DisplaySignOnMessage( void )
{
	if( fIsFat32World )
	{
		printf(szSignOn,SHORTVERSION,BUILDVERSION,COPYRIGHT);
	}
	else
	{
		printf(szSignOn,SHORTVERSION622,BUILDVERSION,COPYRIGHT);
	}
}

/*
	$DOCPROC:meminit

	Initialize the MEMFIXED RAM disk.

	On Entry
		cmdline	the cmdline from device= in CONFIG.SYS
		*endseg	paragraph of code+data. adusted for conventional disk.
		drv		drive # assigned from DOS (0=A, 1=B, etc..)
	On Exit
		bpb		updated with accurate BPB for RAM disk
		*endseg	updated if conventional memory RAM disk

	Returns 1 if worked, 0 if failed.
*/
int meminit(struct BPB bpb[], char far *cmdline, unsigned *endseg, int drv)
{
	int		nSecPerKB;

	//  Determine if we are operating in a DOS 7.1 enviornment
	//
	fIsFat32World = dlIsFat32World();

	/* display sign-on */
	DisplaySignOnMessage();

	/* validate parameters */
	ASSERT(bpb);

	/* determine switches, and sizes suggested by user */
	#if BUILTIN
		cmdline = szBuiltinCmdline;
	#endif
	if(!ParseCmdline(cmdline))
		return 0;

	/* allow up to 512K disks */
	if(nDiskSizeK > 512)
		nDiskSizeK = 512;

	/* limit 8Meg for 128-byte sectors, 16Meg for 256 byte sectors */
	if(nDiskSizeK >= 8*1024 && nBytesPerSector == 128)
		nBytesPerSector = 256;
	if(nDiskSizeK >= 16*1024 && nBytesPerSector == 256)
		nBytesPerSector = 512;

	/* set total # of sectors */
	nSecPerKB = 1024 / nBytesPerSector;
	nTotalSecs = nDiskSizeK * nSecPerKB;

	/* actually format the media */
	if(!FormatRAMDisk(nBytesPerSector, nRootEntries, nTotalSecs))
	{
		printf(szErrFormat);
		return 0;
	}

	/* show installed message */
	printf(szInstalled, nDiskSizeK, (bStatic ? szStatic : ""), drv+'A',
		nStartSeg);

	/* tell memdisk about this BPB */
	GetDiskBPB(bpb);

#if !BUILTIN
	/* claim DOS memory if in conventional */
	if(nStartSeg < memtop)
		*endseg = nStartSeg + (nDiskSizeK * 0x40);	/* 40h paras each 1K */
#endif

	/* one disk */
	return 1;
}

/*
	$DOCPROC:memread

   Read one or more sectors from the RAM disk into a buffer specified
	by DOS.

	On Entry
	   offset   The offset (in bytes) into the disk where the read
					starts.
	   len      The length of the read in bytes. The length cannot
					exceed 64KB - 1 sector.
	   buffer   Pointer to the buffer to read into.

	On Exit
	   Returns TRUE(non-zero) if worked, FALSE(zero) if failed.
*/
bool memread(long offset, unsigned len, char far *buffer)
{
	#if DEBUG > 1
		printf("memread(%lx, %x, %lp)\n",offset,len,buffer);
	#endif

	/* validate parameters */
	ASSERT(len);
	ASSERT(buffer);
	ASSERT(offset < 0x80000L);		/* disks are always <= 512K */
	ASSERT(nStartSeg);

	/* copy the memory */
	fmemcpy(buffer, AddPtr(nStartSeg, offset), len);

	/* read always works */
	return TRUE;
}


/*
	$DOCPROC:memwrite

   Write one or more sectors to the RAM disk from a buffer specified
	by DOS.

	On Entry
	   offset   The offset (in bytes) into the disk where the write
					starts.
	   len      The length of the write in bytes. The length cannot
					exceed 64KB - 1 sector.
	   buffer   Pointer to the buffer to write from.

	On Exit
	   Returns TRUE(non-zero) if worked, FALSE(zero) if failed.
*/
bool memwrite(long offset, unsigned len, char far *buffer)
{
	/* display write address in DEBUG mode */
	#if DEBUG > 1
		printf("memwrite(%lx, %x, %lp)\n",offset,len,buffer);
	#endif

	/* validate parameters */
	ASSERT(len);
	ASSERT(buffer);
	ASSERT(nStartSeg);

	/* write to memory */
	fmemcpy(AddPtr(nStartSeg, offset), buffer, len);

	/* writes always work */
	return TRUE;
}

/*
	$DOCPROC:ParseCmdline

	Parse the command line arguments. Parses both switches and
	arguments as follows:

	Syntax:
	DEVICE=MEMFIXED.SYS [DiskSize [SectorSize [NumEntries]]] [/F#] [/S]
                      ^
*/
bool ParseCmdline(char far *lpCmdline)
{
	char	c;
	int	nArgCount = 0;

	/* validate parameter */
	ASSERT(lpCmdline);

	/* look for command line arguments */
	while(1)
	{
		/* skip whitespace to next switch */
		lpCmdline = MemSkipWhite(lpCmdline);
		c = *lpCmdline;

		/* at end of line? */
		if(c < ' ')
			break;

		/* cmdline options (switches) */
		if(c == '/')
		{
			/* get switch char */
			++lpCmdline;					/* skip '/' */
			c = MemToUpper(*lpCmdline++);

			switch(c)
			{
				/* /F# RAM disk to be placed in extended memory */
				case 'F':
					nStartSeg = 0;
					while(MemIsXDigit(*lpCmdline))
					{
						nStartSeg = nStartSeg * 0x10 + MemXDigitVal(*lpCmdline);
						++lpCmdline;
					}
					break;

				/* /S RAM disk should remain static, if possible */
				case 'S':
					bStatic = TRUE;
					break;

				default:
					printf(szErrBadOpt,c);
					break;
			}
		}

		/* must be a cmdline argument */
		else
		{
			switch(nArgCount)
			{
				/* arg 0 is disk size */
				case 0:
					nDiskSizeK = MinMax(lpCmdline, 4, 512);
					break;

				/* arg 1 is sector size */
				case 1:
					nBytesPerSector = MinMax(lpCmdline, 128, 512);
					if(nBytesPerSector != 128 && nBytesPerSector != 256)
						nBytesPerSector = 512;
					break;

				/* arg 2 is root entries */
				case 2:
					nRootEntries = MinMax(lpCmdline, 4, 512);
					break;

				case 3:
					printf("Too many arguments!\n");
					return FALSE;
			}

			/* skip the argument */
			lpCmdline = MemSkipArg(lpCmdline);

			/* one more argument */
			++nArgCount;
		}

	} /* end of while */

	return TRUE;
}

#if DEBUG
/*
	$DOCPROC:AssertFail

	When an assertion fires, this function is called. Only defined if
	DEBUG is defined.
*/
void PASCAL AssertFail(int nFile, int nLine)
{
	panic("MEMFIXED Internal Error %d.%d!\n",nFile,nLine);
}
#endif	/* DEBUG */


#if UNIT_TEST

int main(void)
{
	static char szUnitTestCmdline[] = "32 /FD000 /S";

	/* return error code if failed. messages are displayed by test. */
	if(!RamDiskUnitTest(szUnitTestCmdline))
		return 1;

	/* no errors */
	return 0;
}

#endif	/* UNIT_TEST */
