//  $Id: PAGEDLDR.C 1.6 2003/03/20 16:52:50 thomd Exp $
//
//  *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//  Copyright (c) 1991-2003 Datalight, Inc.
//  All Rights Reserved
//
//  Datalight, Inc. is a Washington corporation located at:
//
//  21520 30th Dr SE, Suite 110,          Tel: +1.425.951.8086
//  Bothell, WA  98021                    Fax: +1.425.951.8094
//  USA                                   Http://www.datalight.com
//
//  This software, including without limitation all source code and
//  documentation, is the confidential, trade secret property of
//  Datalight, Inc., and is protected under the copyright laws of
//  the United States and other jurisdictions.  Portions of the
//  software may also be subject to one or more the following US
//  patents: US#5860082, US#6260156.
//
//  In addition to civil penalties for infringement of copyright
//  under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//  penalties for violation of (a) the restrictions on circumvention
//  of copyright protection systems found in 17 U.S.C. 1201 and
//  (b) the protections for the integrity of copyright management
//  information found in 17 U.S.C. 1202.
//
//  U.S. Government Restricted Rights.  Use, duplication,
//  reproduction, or transfer of this commercial product and
//  accompanying documentation is restricted in accordance with
//  FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//  IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//  PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//  AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//  INC. AND THE LICENSEE ("BINDING AGREEMENT").  YOUR RIGHT, IF ANY,
//  TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//  TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//  USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//  TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT.  CONTACT
//  DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//  THIS SOFTWARE IN ERROR.
//
/********************************************************************
	$DOCMODULE:PAGEDLDR.C

	Copyright (c) 1991-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary.

	This module works together with LDRBEXT.ASM to copy ROM-DOS from
	paged memory to RAM. It is useful for those systems that don't
	have the real-mode address space to have all of the ROM-DOS kernel
   (48K) in ROM or Flash, but still need to load ROM-DOS from a
	memory disk.

	Three modules must be linked, in the following order:
	$(LINK) $(LFLAGS) ldrbext+pagedldr+sc400pag,sc400pag,,memutil

	$DOCHISTORY
	07/23/1998 dg  Converted to Solid Code
	11/20/1991 dg  Created by Drew Gislason
********************************************************************/
#include <dos.h>
#include "compiler.h"
#include "memdisk.h"

/* a far ptr to load address, normally 70:0 (0x700000L) */
#if UNIT_TEST
 #define LOAD_PTR	 0x80000000L	/* 8000:0 for testing purposes */
#else
 #define LOAD_PTR	 0x700000L		/* 70:0 */
#endif

/* prototypes */
void bputc(char c);
void bputs(char far *s);
void bputh(unsigned val);
void bputb(unsigned char val);
void bputlp(void far *lpAddress);
void bdump(void far * _lpAddress);
void waitkey(void);

/*
	$DOCPROC:IsValidBPB

	Simple routine to validate the BPB. Not very smart (yet).
	Returns TRUE if valid, FALSE if not.
*/
bool IsValidBPB(struct BPB far *lpBPB)
{
	/* BPB must look reasonable */
	if( ((lpBPB->MediaDescrip & 0xF0) != 0xF0) ||
				!lpBPB->BytesPerSector || (lpBPB->BytesPerSector&0x7F))
		return FALSE;

	/* checks out OK */
	return TRUE;
}

/*
	$DOCPROC:GetWindowPtr

	Sets the window to the page at the current starting offset.

	On Entry
		lStart	Starting offset into the disk.
*/
void far * GetWindowPtr(long lStart)
{
	unsigned nPage;
	unsigned	nOffset;

	/* determine page this starts on */
	nPage = (unsigned)_ldiv(lStart, lPagedDiskWindowSize);
	nOffset = (unsigned)_lmod(lStart, lPagedDiskWindowSize);

	/* set hardware to that page */
	set_page(nPage);

	/* return pointer to that memory */
	return (lpPagedDiskWindow + nOffset);
}

/*
	$DOCPROC:FindRomDos

	Find the ROM-DOS.SYS file.
	On Entry
		lpStart	beginning of the disk.
	On Exit
		lpStart	beginning of the ROM-DOS.SYS file.
		lpSize	size of ROM-DOS.SYS file.

	Returns TRUE if found, FALSE if failed.
*/
bool FindRomDos(long far *lpStart, unsigned far *lpSize)
{
	unsigned dirEntries;	/* # of entries in root dir */
	unsigned dirStart;	/* start of root dir in bytes */
	unsigned clusSize;	/* size of a cluster */
	long		startFile;	/* start of file in bytes */
	unsigned i;
	struct DIR far * lpDir;
	struct BPB far *lpBPB;
	static char romDosName[] = "ROM-DOS SYS";

	/* get pointer to BPB */
	lpBPB = GetWindowPtr(*lpStart+11);

	/* is BPB valid? */
	if(!IsValidBPB(lpBPB))
		return FALSE;

	/* determine start of disk */
	dirStart = (lpBPB->Reserved + lpBPB->SecPerFAT * lpBPB->NumbFats) *
						lpBPB->BytesPerSector;
	dirEntries = lpBPB->MaxDirs;
	clusSize = lpBPB->SecPerCluster * lpBPB->BytesPerSector;

	#if (DEBUG || UNIT_TEST)
		bputs("\nbps "); bputh(lpBPB->BytesPerSector);
		bputs("\ndir "); bputh(dirStart);
		bputs("\nent "); bputh(dirEntries);
		bputs("\nbpc "); bputh(clusSize);
		waitkey();
		bputc('\n');
	#endif

	/* look through root directory for ROM-DOS.SYS */
	for(i=0; i<dirEntries; ++i)
	{
		lpDir = GetWindowPtr(*lpStart + dirStart + i*sizeof(struct DIR));
		if(fmemcmp(romDosName, lpDir->name, sizeof(romDosName)-1) == 0)
			break;
	}

	/* ROM-DOS.SYS not found! */
	if(i == dirEntries)
		return FALSE;

	/* determine start of ROM-DOS.SYS */
	startFile = _lmulu((long)(lpDir->start-2), clusSize) +
						dirStart + dirEntries * sizeof(struct DIR);

	/* add in actual start of ROM-DOS.SYS file */
	*lpStart += startFile;
	*lpSize = (unsigned)lpDir->size;

	/* let user know the file size is too big! */
	if(lpDir->size >= 64L*1024)
	{
		bputs("Error! ROM-DOS size greater than 64K!\n");
		return FALSE;
	}

	return TRUE;
}


/*
	$DOCPROC:Loader

	This is the main workhorse function. It uses the paging mechanism
	to load the ROM-DOS.SYS from ROM or Flash into RAM.
*/
void Loader(void)
{
	unsigned nPageBytes;		/* bytes to copy in this page */
	unsigned fileSize;
	uchar far *lpDstAddr;
	uchar far *lpSrcAddr;
	int (far *ROM_DOS)(void);
	long 		lImageOffset;
	long		lPageOffset;

	/* sign on message */
   bputs("Loading ROM-DOS.SYS from paged disk...");

	/* initialize the paging software */
	init_page();

	/* where in RAM will we load ROM-DOS? */
   lpDstAddr = (uchar far *)LOAD_PTR;

	/* Find ROM-DOS.SYS on the ROM-DISK */
	lImageOffset = 0L;
	if(!FindRomDos(&lImageOffset, &fileSize))
	{
	   bputs("not found! booting from floppy!\n");
	   return;
	}

	/*
		Loop through, using as many pages as required to get all of
		the image.
	*/
	while(fileSize)
	{
		/* get pointer to the page (maybe offset into the page) */
		lpSrcAddr = GetWindowPtr(lImageOffset);

		/* The number of bytes to copy from this page */
		lPageOffset = _lmod(lImageOffset, lPagedDiskWindowSize);
		nPageBytes = (unsigned)(lPagedDiskWindowSize - lPageOffset);
		if(nPageBytes > fileSize)
			nPageBytes = fileSize;

		#if (DEBUG || UNIT_TEST)
	  		bputs("\nCopying "); bputh(nPageBytes);
			bputs(" from "); bputlp(lpSrcAddr);
			bputs(" to "); bputlp(lpDstAddr); bputc('\n');
			bdump(lpSrcAddr);
		#endif

		/* set the page and copy data */
		fmemcpy(lpDstAddr, lpSrcAddr, nPageBytes);

		/* update count and pointers */
		lImageOffset += nPageBytes;
		fileSize    -= nPageBytes;
		lpDstAddr    = AddNormalize(lpDstAddr, nPageBytes);
	}

	/* determine */
	ROM_DOS = (void far *)LOAD_PTR;

	#if UNIT_TEST
		/* don't actually jump, just act like we are */
		bputs("\nBooting ROM-DOS from Paged Disk...\n");

		bputlp(ROM_DOS);
		waitkey();

	#else
		#if DEBUG
			bputlp(ROM_DOS);
			waitkey();
		#endif

		/* far jump to the ROM-DOS code */
		(*ROM_DOS)();
	#endif
}

/*
	Displays one character using BIOS screen calls.
	All other bput... routines are built on this function.
*/
void bputc(char c)
{
	/* handle cr/lf pairs in strings */
	if(c == '\n')
		bputc('\r');

	/* output the character using function 0EH of BIOS Int 10H */
	ASM	mov	AL, c
	ASM	mov	AH, 0x0E
	ASM	mov	BX, 7
	ASM	int	0x10
}

/*
	Display an asciiz string using BIOS screen calls.
	Note: must be a FAR pointer to be XIP, because SS != DS.
*/
void bputs(char far *s)
{
	while(*s)
		bputc(*s++);
}

#if (DEBUG || UNIT_TEST)

char hexDigits[] = "0123456789ABCDEF";

/*
	Display a 16-bit hex number via BIOS screen calls.
*/
void bputh(unsigned val)
{
	int i;

	/* display all 4 digits as ascii */
	for(i=12; i>=0; i-=4)
		bputc(hexDigits[(val >> i) & 0xF]);
}

/*
	Display an 8-bit hex number via BIOS screen calls.
*/
void bputb(unsigned char val)
{
	int i;

	/* display all 2 digits as ascii */
	for(i=4; i>=0; i-=4)
		bputc(hexDigits[(val >> i) & 0xF]);
}

/*
	Display a 32-bit number as a pointer.
*/
void bputlp(void far *lpAddress)
{
	unsigned seg, off;

	/* convert to segment:offset */
	seg = FP_SEG(lpAddress);
	off = FP_OFF(lpAddress);

	/* display as such */
	bputh(seg);
	bputc(':');
	bputh(off);
}

/*
	Dump one line (16 bytes) worth of bytes.
*/
void bdump(void far * _lpAddress)
{
	int i;
	unsigned char far * lpAddress = _lpAddress;
	static char spaces[] = "  ";

	/* display the address of these bytes */
	bputlp(lpAddress);
	bputs(spaces);

	/* display the bytes in hex form */
	for(i=0; i<16; ++i)
	{
		/* show a vertical bar in the middle for readability */
		if(i == 8)
		{
			bputc('|');
			bputc(' ');
		}

		/* display the byte */
		bputb(lpAddress[i]);
		bputc(' ');
	}
	bputs(spaces);

	/* display the bytes in ascii form */
	for(i = 0; i < 16; i++)
	{
		/* ascii chars are displayed as-is */
		if(lpAddress[i] >= ' ' && lpAddress[i] <= '~')
			bputc(lpAddress[i]);

		/* non-printable chars are displayed as dots */
		else
			bputc('.');
	}

	/* next line */
	bputc('\n');
}

/*
	Wait for a key from the user. Uses BIOS Int 16H.
*/
void waitkey(void)
{
	ASM	xor ax, ax
	ASM	int 16h
}
#endif

#if UNIT_TEST
void main(void)
{
	char far * lpWorld = "world";

	/* unit test debugging output tools */
	bputc('a');
	bputs(" Hello ");
	bputs(lpWorld);
	bputh(0x1234);
	bputb(0xAC);
	bputlp(lpWorld);
	bdump(lpWorld);

	/* test the loader */
	Loader();
}
#endif
