//  $Id: CMDCMDS.C 1.5 2003/03/20 16:59:36 thomd Exp $
//
//  *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//  Copyright (c) 1992-2003 Datalight, Inc.
//  All Rights Reserved
//
//  Datalight, Inc. is a Washington corporation located at:
//
//  21520 30th Dr SE, Suite 110,          Tel: +1.425.951.8086
//  Bothell, WA  98021                    Fax: +1.425.951.8094
//  USA                                   Http://www.datalight.com
//
//  This software, including without limitation all source code and
//  documentation, is the confidential, trade secret property of
//  Datalight, Inc., and is protected under the copyright laws of
//  the United States and other jurisdictions.  Portions of the
//  software may also be subject to one or more the following US
//  patents: US#5860082, US#6260156.
//
//  In addition to civil penalties for infringement of copyright
//  under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//  penalties for violation of (a) the restrictions on circumvention
//  of copyright protection systems found in 17 U.S.C. 1201 and
//  (b) the protections for the integrity of copyright management
//  information found in 17 U.S.C. 1202.
//
//  U.S. Government Restricted Rights.  Use, duplication,
//  reproduction, or transfer of this commercial product and
//  accompanying documentation is restricted in accordance with
//  FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//  IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//  PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//  AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//  INC. AND THE LICENSEE ("BINDING AGREEMENT").  YOUR RIGHT, IF ANY,
//  TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//  TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//  USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//  TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT.  CONTACT
//  DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//  THIS SOFTWARE IN ERROR.
//
/*********************************************************

   CMDCMDS.C  created October 1992

   Copyright (c) 1992-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary

   All commands for the miniCOMMAND processor are coded
   in this file.

**********************************************************/
#include "minicmd.h"
#include "dosver.h"

#if CMD_GOTO
void do_goto(char *args)
{
	extern char command[82];

	/* goto only works in batch files */
	if(!In_Batch)
		return;

	/* search from the beginning of the batch file for the label */
	Batch_pos = 0;
	while(In_Batch)
	{
		get_file_line(command);
		if(command[2]==':' && labelcmp(command+3,args)==0)
			break;
	}

	/* label not found, restore batch position */
	if(!In_Batch)
		puts("Label not found\n");
}

#endif			/* CMD_GOTO */

#if CMD_GOTO || CMD_IF

int labelcmp(char *str1, char *str2)
{
	while(toupper(*str1) == toupper(*str2))
	{
		if(*str1 <= ' ')
			break;
		++str1, ++str2;
	}
	if(*str1 <= ' ' && *str2 <= ' ')
		return 0;
	return *str1 < *str2 ? -1 : 1;
}

#endif			/* CMD_GOTO OR CMD_IF*/

#if CMD_IF
void do_if(char *args)
{
	extern char command[82];
	char *filename = 0;
	char *action = 0;

	/* next word MUST be exist */
	if(labelcmp(args, "exist") != 0)
	{
SyntaxErr:
		puts("Syntax error\n");
		In_Batch = 0;
		return;
	}

	/* check for file */
	filename = strchr(args, ' ');
	if(filename)
	{
		filename = skipwhite(filename);
		action = strchr(filename, ' ');
		if(action)
		{
			*action = 0;
			action = skipwhite(action+1);
		}
	}
	if(!filename || !action)
		goto SyntaxErr;
	if(exists(filename))
	{
		strcpy(command+2,action);
		command[1] = strlen(action)+1;
		strcat(command+2,"\n");
		make_cmd();
		do_one_command();
	}
}

/*
 * exists
 *
 * This function is checks if a file or subdir exists.
 * Return
 *	True if file or subdir exists
 *	False if it does not
 */
exists(char *file)
{
	struct FFBLK ffblk;
	if(findfirst(file, &ffblk, 0xFF) == 0)
		return 1;
	return 0;
}

#endif			/* CMD_IF */


#if CMD_SET
char far *env;

/*
 * do_set
 *
 * This function is the miniCOMMAND SET command.
 *
 * Return
 * 	None
 */
void do_set(char *args)
{
	char *end;
	unsigned i;

	env = MK_FP(psp->Env_Seg, 0);
	if(strlen(args) < 2)
	{
		while(*env)			/* Display Environment */
		{
			while(*env)
				putc(*env++);
			puts("\n");
			++env;
		}
	}
	else 					/* Modify Envrionment */
	{
		end = strchr(args, '=');

		for(i=0; i<end-args; ++i)			/* Make sure it's upper Case */
			args[i] = toupper(args[i]);

		if(strlen(end) > 1)
			add_env(args);
		else
			del_env(args);

	}
}

/*
 * del_env
 *
 * This function is a utility routine for the do_set
 * function.  It deletes the environment.
 *
 * Return
 * 	None
 */
void del_env(char *str)
{
	char far *var, far *end_of_var, *p;
	unsigned size;

	var = get_env_full(str);

	if(var != NULL)
	{
		for(end_of_var=var; *end_of_var; ++end_of_var)
			;
		size = env_avail_size() - FP_OFF(end_of_var);
		far_memcpy(var, end_of_var+1, size);
	}
}

/*
 * add_env
 *
 * This function is a utility routine for the do_set function.  It adds
 * a string to the environment.
 *
 * Return
 * 	None
 */
void add_env(char *str)
{
	int len = strlen(str), left;

	left = env_avail_size() - (env_size() + len);

	if(left < 3)
		error(6);
	else
	{
		del_env(str);
		str[strlen(str) + 1] = '\0';			/* Add second Zero */
		far_memcpy(env + env_size(), str, len + 2);
	}
}


/*
 * env_size
 *
 * This function is a utility routine for the do_set function.  It determines
 * the size of the environment.
 *
 * Return
 * 	The environment size in bytes
 */
unsigned env_size(void)
{
	unsigned count=0;

	while(env[count])
	{
		while(env[count])
			++count;
		++count;
	}
	return(count);
}



/*
 * env_avail_size
 *
 * This function is a utility routine for the do_set function.  It
 * gets the maximum size of the environment.
 *
 * Return
 * 	The maximum environment size in bytes
 */
unsigned env_avail_size(void)
{
	struct MCB far *env_mcb = MK_FP(psp->Env_Seg-1, 0);
	return (env_mcb->size * 16);
}


#endif			 /* CMD_SET */


char Echo = TRUE;		/* This must be defined even if do_echo is not! */

#if CMD_ECHO

char *get_token(char *tok, char *string);

/*
 * do_echo
 *
 * This function is the miniCOMMAND ECHO command.
 *
 * Return
 * 	None
 */
void do_echo(char *args)
{
	if(get_token("ON", args))
		Echo = TRUE;
	else if(get_token("OFF", args))
		Echo = FALSE;
	else if(strlen(args) > 1)
		puts(args);
	else
	{
		puts("ECHO is ");
		puts(Echo ? "ON" : "OFF");
	}
}


/*
 * get_token
 *
 * This function gets a character token out of a string.
 *
 * Return
 * 	String 	- if token found.
 *	NULL 	- If token not found.
 */
char *get_token(char *tok, char *string)
{
	while(iswhite(*string))
		++string;
	while(*tok)
	{
		if(toupper(*string) != *tok)
			return (NULL);
		++tok;
		++string;
	}
	return (string);
}

#endif		/* CMD_ECHO */



#if CMD_PAUSE

/*
 * do_pause
 *
 * This function is the miniCOMMAND PAUSE command.
 *
 * Return
 * 	None
 */
void do_pause(char *args)
{
	char str[10];
	puts(args);
	puts("Strike a key when ready . . .");
	getc(2);
}



#endif

#if CMD_DEL


/*
 * do_del
 *
 * This function is the miniCOMMAND DELETE (DEL) command.
 *
 * Return
 * 	None
 */
void do_del(char *args)
{
	dosdelete(args);
	if(_errno)
		error(3);			/* File Not found */
}

#endif CMD_DEL


#if CMD_COPY

#define BUFSIZE (4096)

void do_copy(char *args)
{

	char *File1, *File2;

	File1 = args;
	File2 = strchr(args, ' ');
	if(File2 == NULL)
		puts("Second file not specified!\n");
	else
	{
		*File2 = '\0';
		++File2;
		copy_1_file(File1, File2);
	}
}

void copy_1_file(char *File1, char *File2)
{
	int input, output;
	unsigned count;
	char far *buffer;

	input = dosopen(File1, 0);
	if(input == -1)
	{
		error(7);
		return;
	}

	output = doscreate(File2);
	if(output == -1)
	{
		error(7);
		return;
	}

	buffer = dosmalloc(BUFSIZE/16);

	do
	{
		count = dosread(input, buffer, BUFSIZE);
		if(count)
			doswrite(output, buffer, count);
	}
	while(count);
	dosclose(input);
	dosclose(output);
	dosfree(buffer);
}



#endif 	/* CMD_COPY */




#if CMD_DIR

/*
 * do_dir
 *
 * This function is the miniCOMMAND DIR command.
 *
 * Return
 * 	None
 */
void do_dir(char *path)
{
	struct FFBLK ffblk;
	int ret, len = strlen(path);

	if(len == 0 || (len == 2 && path[1] == ':'))
	{
		strcat(path, "*.*");
		ret = findfirst(path, &ffblk, 0x10);
	}
	else
	{
		ret = findfirst(path, &ffblk, 0x10);
		if((ffblk.attrib & 0x10) && ffblk.reserved[1] != '?')
		{
			strcat(path, "\\*.*");
			ret = findfirst(path, &ffblk, 0xFF);
		}
	}

	if(ret)
	{
		error(3);
		return;
	}

	while(!ret)
	{
		puts(ffblk.name);
		if(ffblk.attrib & A_DIR)
			puts("\t<DIR>");
		puts("\n");
		ret = findnext(&ffblk);
	}
}

#endif 	/*	CMD_DIR */



#if CMD_TYPE


/*
 * do_type
 *
 * This function is the miniCOMMAND TYPE command.
 *
 * Return
 * 	None
 */
void do_type(char *args)
{
	copy_1_file(args, "CON");
}


#endif CMD_TYPE


#if CMD_CD

/*
 * do_chdir
 *
 * This function is the miniCOMMAND CHDIR command.
 *
 * Return
 * 	None
 */
void do_chdir(char *args)
{
	int drive;

	if(strlen(args) == 0)
		show_cwd(args, 0);
	else if((args[1] == ':' && args[2] == '\0'))
	{
		drive = toupper(args[0]) - 'A' + 1;
		show_cwd(args, drive);
	}
	else
	{
		if(set_cwd(args) == -1)
			error(8);
	}
}

void show_cwd(char *args, unsigned drive)
{
		*args++ = '\\';
		get_cwd(drive, args);
		puts(--args);
		puts("\n");
}

#endif		/* CMD_CD */


#if CMD_EXIT

/*
 * do_exit
 *
 * This function performs an exit if miniCOMMAND was not run
 * with the /Perminate switch.
 *
 * Return
 * 	None
 */
#pragma argsused
void do_exit(char *parms)
{
	extern int exit_allowed;
	if(exit_allowed)
		exit();
}

#endif 	/* CMD_EXIT */


#if CMD_VER

/*
 * do_ver
 *
 * This function is the miniCOMMAND Version command.  It
 * displays the miniCOMMAND Version.
 *
 * Return
 * 	None
 */
#pragma argsused
void do_ver(char *parms)
{
	puts("\nminiCOMMAND ");
	puts(FULLVERSION);
	puts("\n");
	puts(COPYRIGHT);
	puts("\n");

	/* show internal version as well */
	if(strchr(args, '/'))
	{
		puts("\nROM-DOS is in ");
		switch(get_dos_mem_position())
		{
			case 0x10: puts("HMA"); break;
			case 0x8:  puts("ROM"); break;
			default:   puts("RAM"); break;
		}
		puts("\n");
	}
}

#endif		/* CMD_VER 	*/
