//
//   $Id: MALLOC.C 1.2 2002/10/16 16:01:26 peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*********************************************************

	MALLOC.C

	Copyright (c) 1989-1996 by Datalight, Inc.
	All Rights Reserved.

	This file contains the memory allocation routines for
	heap management.

	The heap is a single linked list of memory blocks.
	Each block is idenitified by an unsigned number that is
	the number of bytes in the block. If the number is odd
	then it is NOT allocated, even means that the block is
	allocated.  Obviously then, allocation is only done in
	even size numbers.

**********************************************************/


// Malloc only for Borland Large Model
//
#ifdef __TURBOC__
	#if !(defined( __LARGE__ ))
   	#if (sizeof(void *) == sizeof(long))
    		#define __BIGBCC__	1
		#else
			#define __BIGBCC__	0
		#endif
	#else
		#if( __LARGE__ || __HUGE__ || __COMPACT__ )
			#define __BIGBCC__	1
		#else
			#define __BIGBCC__	0
		#endif
	#endif
#else
	#define __BIGBCC__			0
#endif

#if __BIGBCC__		// if BCC far data...

#include <mem.h>
#include "compiler.h"

#define TESTALLOC	0			// 0 for normal use or "1" for Testing!


#define MK_FP(seg,off) \
	((void far *)(((unsigned long)(seg)<<16) | (unsigned)(off)))
#define FP_SEG(farptr)	((unsigned)((unsigned long)(farptr) >> 16))
#define FP_OFF(farptr)	((unsigned)(farptr))

void far_free(void far * id);
void far * far_malloc(unsigned long bytes);
void  free(void * id);
static void far * normalize(void far * p);
static void far *GetAllDOSMemory(unsigned *paras);


/*
	A memory block.  Effectively a linked list.
*/
typedef struct
{
	unsigned long size;
	char mem[2];
} BLK;


#define SMALLEST 12		/* smallest chunk left "free" */
#define FREE(p)    		((unsigned)((p)->size) & 1)
#define SIZE(p)			((p)->size & 0xFFFFFFFEL)
#define END_OF_LIST(p)  (FP_SEG(p) > FP_SEG(heap_end) || (void far *)(p) == heap_end)


static BLK far * heap_start;
static void far * heap_end;

/*
	Determine the next memory block.  Always returns a
	normalized pointer.  Can handle a "huge" piece of
	memory.  No error checking.
*/
static BLK far * NEXT(BLK far * p)
{
	unsigned long addr;

	addr = ((long)FP_SEG(p)<<4) + FP_OFF(p);
	addr += SIZE(p) + sizeof(long);	/* account for memory and "size" field */

	return (BLK far *)MK_FP((unsigned)(addr>>4), ((unsigned)addr&0xF));
}


/*
	Convert a near pointer into a "normalized" far pointer.
	Always works.
*/
static void far * normalize(void far * p)
{
	unsigned seg, off;

	seg = FP_SEG(p);
	off = FP_OFF(p);
	seg = seg + (off>>4);
	off = off & 0xF;
	return MK_FP(seg,off);
}


/*
	Allocate a number of bytes.  returns a pointer to that memory
	or NULL if none available.  Note: a near pointer will be
	returned in SMALL data models, a far pointer will be returned
	in LARGE data models.
*/
void * malloc(unsigned bytes)
{
	void far * fm;
	void * mem = NULL;

	fm = far_malloc((unsigned long)bytes);
	if(fm)
	{
#if sizeof(void *) != sizeof( void far *)
		// This code is for large model only!!!

	This is an Error!!!

	Large Model Only!!!
#endif

		mem = fm;
	}

	return mem;
}

/*
	Allocate the requested # of bytes from the far memory
	pool.  The near and far pool are synonymous, only the
	type of pointer returned is different.
*/
void far * far_malloc(unsigned long bytes)
{
	/*
		[ locals ]
	*/
	static int first_time = 1;
	BLK far * p;
	BLK far * next;
	unsigned long leftover;
	unsigned long addr;
	unsigned paras;


	/*
		[ code ]
	*/

	/* Want no bytes?  You got no bytes */
	if(!bytes)
		return (NULL);

	/* initialize the memory system */
	if(first_time)
	{
		first_time = 0;

		/* Determine start and end of heap */
		heap_start = GetAllDOSMemory(&paras);
		heap_end = MK_FP(FP_SEG(heap_start) + paras, 0);

		/* all memory is free : odd = free, even = allocated */
		heap_start->size = (16L * (FP_SEG(heap_start) - FP_SEG(heap_end)) - sizeof(long)) | 1;
	}

	/* allocate only even # of bytes */
	if(bytes & 1)
		++bytes;

	/* try to allocate requested bytes of memory */
	p = heap_start;
	while(!END_OF_LIST(p))
	{
		if(FREE(p) && SIZE(p) >= bytes)
		{
			leftover = SIZE(p) - bytes;
			if(leftover > SMALLEST)	      /* break up a bigger block into 2 */
			{
				addr = ((unsigned long)FP_SEG(p) << 4) + FP_OFF(p);
				addr+= bytes + sizeof(long);
				next = (BLK far *)MK_FP((unsigned)(addr>>4),(unsigned)addr&0xF);
				next->size = (leftover - sizeof(long)) | 1 ;
			}
			else
				bytes = p->size & 0xFFFFFFFEL;	/* allocate it */

			p->size = bytes;
			return (p->mem);
		}

		/* not a free block or not big enough, go on to next */
		else
			p = NEXT(p);
	}

	/* cannot allocated requested memory, fail */
	return (NULL);
}


void free(void * id)
{
	far_free((void far *)id);
}


/*
	Free up some memory allocated by malloc.  Verifies that
	the block is valid before it will free the memory.  Does
	NOT assume "id" pointer is normalized.
*/
void far_free(void far * id)
{
	void far * pid;
	BLK far * p;
	BLK far * prev;

	pid = normalize(id);
	prev = p = heap_start;

	while(!END_OF_LIST(p))
	{
		if(pid == normalize(p->mem))
		{
			/* Found it */
			if(FREE(prev) && p != heap_start)
				prev->size += SIZE(p) + sizeof(long);
			else
				prev = p;
			if(FREE(NEXT(p)))
				prev->size += SIZE(NEXT(p)) + sizeof(long);
			prev->size |= 1;		/* mark block as free */
			return;
		}

		/* not found, look at the next one */
		else
		{
			prev = p;
			p = NEXT(p);
		}
	} /* done searching list */
}


/*
	Allocate a number of items.  Zero the space.
*/
void * calloc(unsigned items, unsigned size)
{
	void * buff;

	buff = malloc(items * size);
	if(buff)
		memset(buff, 0, items * size);
	return buff;
}



/*
 * Get all available DOS memory!
 */
static void far *GetAllDOSMemory(unsigned *paras)
{
	unsigned segment, numParas;

	ASM	mov	ax, 0x4800
	ASM	mov	bx, 0xFFFF
	ASM	int	0x21
	ASM	mov	numParas, bx

	*paras = numParas;

	ASM	mov	ax, 0x4800
	ASM	int	0x21
	ASM	mov	segment, ax
	ASM	jnc	Done

Failed:
	*paras = 0;
	segment = 0;

Done:
	return (MK_FP(segment, 0));
}

#if TESTALLOC
#include <stdio.h>
#include <stdlib.h>
/*
	FOR TESTING PURPOSES ONLY!!
	chk_mem is used to verify if the memory system is not currupted
	It produces a list of memory blocks
*/
void showheap(void)
{
	BLK far * p = heap_start;

	while(!END_OF_LIST(p))
	{
		printf("%05lX Bytes %s at %04X:%04X\n", p->size & (long)0xFFFFFFFE,
			FREE(p) ? "FREE     " : "ALLOCATED", FP_SEG(p), FP_OFF(p));
		p = NEXT(p);
	}
}

/*
	FOR TESTING PURPOSES ONLY!!
	This main tests the malloc and free.  Set TESTALLOC to 1
	to include main().
*/
void main(void)
{
	char *p1, *p2, *p3, *p4;

	unsigned paras;

	GetAllDOSMemory(&paras);			// Allocate some mem

	GetAllDOSMemory(&paras);			// Allocate some mem

	GetAllDOSMemory(&paras);			// Allocate some mem

	showheap();

	printf("test allocation : p1 = malloc(2000);\n");
	p1 = malloc(2000);
	printf("  p1 = %p\n",p1);
	showheap();

	printf("test allocation of odd byte sizes : p2 = malloc(19);\n");
	p2 = malloc(19);
	printf("  p2 = %p\n",p2);
	showheap();

	printf("test freeing : free(p1);\n");
	free(p1);
	showheap();

	printf("test allocating some in the middle : p1 = malloc(1000);\n");
	p1 = malloc(1000);
	printf("  p1 = %p\n",p1);
	showheap();

	printf("test allocating too much for 1st found space : p3 = malloc(1000);\n");
	p3 = malloc(1000);
	printf("  p3 = %p\n",p3);
	showheap();

	printf("test freeing of two contiguous areas of memory : free(p1);\n");
	free(p1);
	showheap();

	printf("test 'smallest' alloc_block code : p1 = malloc(1990);\n");
	p1 = malloc(1990);
	printf("  p1 = %p\n",p1);
	showheap();

	printf("test no enough memory : p4 = malloc(64000);\n");
	p4 = malloc((unsigned)64000);
	printf("  p4 = %p\n",p4);
	showheap();

	printf("test free all : free(p1), free(p2), free(p3), free(p4)\n");
	free(p1);
	free(p2);
	free(p3);
	free(p4);
	showheap();
}

#endif			// TESTALLOC
#endif			// __BIGBCC__
