//  $Id: STELNET.C 1.3 2003/05/08 14:17:50 johnb Exp $
//
//  *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//  Copyright (c) 1999-2003 Datalight, Inc.
//  All Rights Reserved
//
//  Datalight, Inc. is a Washington corporation located at:
//
//  21520 30th Dr SE, Suite 110,          Tel: +1.425.951.8086
//  Bothell, WA  98021                    Fax: +1.425.951.8094
//  USA                                   Http://www.datalight.com
//
//  This software, including without limitation all source code and
//  documentation, is the confidential, trade secret property of
//  Datalight, Inc., and is protected under the copyright laws of
//  the United States and other jurisdictions.  Portions of the
//  software may also be subject to one or more the following US
//  patents: US#5860082, US#6260156.
//
//  In addition to civil penalties for infringement of copyright
//  under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//  penalties for violation of (a) the restrictions on circumvention
//  of copyright protection systems found in 17 U.S.C. 1201 and
//  (b) the protections for the integrity of copyright management
//  information found in 17 U.S.C. 1202.
//
//  U.S. Government Restricted Rights.  Use, duplication,
//  reproduction, or transfer of this commercial product and
//  accompanying documentation is restricted in accordance with
//  FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//  IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//  PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//  AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//  INC. AND THE LICENSEE ("BINDING AGREEMENT").  YOUR RIGHT, IF ANY,
//  TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//  TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//  USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//  TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT.  CONTACT
//  DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//  THIS SOFTWARE IN ERROR.
//
// Simple BSD sockets Telnet program

#include <stdio.h>
#include <stdlib.h>
#ifdef _MSC_VER
#include <malloc.h>
#else
#include <mem.h>
#endif
#include <string.h>
#include <conio.h>
#include <errno.h>
#ifdef _Windows
#include <winsock2.h>
#define Errno WSAGetLastError()
#else
#include "socket.h"
#define Errno errno
#endif

// prototypes

char *GetSocketError(int iError);
void Fail(char *pszCall);
void SendData(char *buf,int cnt);

//telnet functions

void InitTelnet(void);
int TeloptStrip(char *buf,int cnt);
void WillOpt(char opt);
void WontOpt(char opt);
void DoOpt(char opt);
void DontOpt(char opt);
void SendOpt(char cmd,char opt);

int iTelnetState;				// telnet state
SOCKET s;


int main(int argc,char *argv[])
{
	int i,len;
	unsigned long ulOn = 1l;
	struct sockaddr sSockAddrLocal,sSockAddrRemote;
	struct sockaddr_in *pAddr = (struct sockaddr_in *)&sSockAddrRemote;
	char pcBuf[100],cKey,*pszHostName = "vp",*pszServiceName = "telnet";
	struct servent *psService;
	struct hostent *psHost;

#ifdef _Windows
	WSADATA 	winsockdata;

	if (WSAStartup(0x0101,&winsockdata) != 0) {
		WSACleanup();
		printf("WinSock V1.1 required\n");
		exit(1);
	}
	printf("Winsock v %u.%u\n",HIBYTE(winsockdata.wHighVersion),LOBYTE(winsockdata.wHighVersion));
#endif

	s = socket(AF_INET,SOCK_STREAM,IPPROTO_TCP);
	if (s == INVALID_SOCKET)
		Fail("socket");
	printf("Socket is %d\n",s);
	memset(&sSockAddrLocal,0,sizeof(sSockAddrLocal));
	memset(&sSockAddrRemote,0,sizeof(sSockAddrRemote));
	sSockAddrLocal.sa_family = AF_INET;
	sSockAddrRemote.sa_family = AF_INET;
	if (bind(s,&sSockAddrLocal,sizeof(sSockAddrLocal)) == SOCKET_ERROR)
		Fail("bind");
	if (argc > 1)
		pszHostName = argv[1];
	if ((psHost = gethostbyname(pszHostName)) == 0) {
		Fail("gethostbyname");
		if ((pAddr->sin_addr.s_addr = inet_addr(argv[1])) == 0)
			pAddr->sin_addr.s_addr = inet_addr("192.168.0.2");
	}
	else
		pAddr->sin_addr.s_addr = *(u_long *)(psHost->h_addr_list[0]);
	if (argc > 2)
		pszServiceName = argv[2];
	if ((pAddr->sin_port = htons((u_short)atoi(pszServiceName))) == 0) {
		if ((psService = getservbyname(pszServiceName,"tcp")) == 0) {
			Fail("getservbyname");
			pAddr->sin_port = htons(23);
		}
		else
			pAddr->sin_port = psService->s_port;
	}
	printf("port= %d\n",ntohs(pAddr->sin_port));
	if (connect(s,&sSockAddrRemote,sizeof(sSockAddrRemote)) == SOCKET_ERROR)
		Fail("connect");
	if (ioctlsocket(s,FIONBIO,&ulOn) == SOCKET_ERROR)
		Fail("ioctlsocket");
	while (1) {
		if (kbhit() != 0) {
			cKey = (char)getch();
			if (cKey == 0x1b)
				break;
			SendData(&cKey,1);
		}
		if ((len = recv(s,pcBuf,sizeof(pcBuf),0)) == SOCKET_ERROR) {
			if (Errno != WSAEWOULDBLOCK) {
				Fail("recv");
				break;
			}
		}
		else if (len == 0) {
			printf("\nPeer closed connection\n");
			break;
		}
		else {
			len = TeloptStrip(pcBuf,len);
			for (i = 0;i < len;++i)
				putchar(pcBuf[i]);
		}
	}
	if (closesocket(s) == SOCKET_ERROR)
		Fail("closesocket");
#ifdef _Windows
	WSACleanup();
#endif
	return 0;
}

/*
Send bytes on s. Returns immediately (non blocking).
If an error occurs, the error message is printed to std out.

Arguments:
	pcBuf - pointer to buffer containing characters to be send.
	iCnt - Number of bytes to send.
*/
void SendData(char *pcBuf, int iCnt)
{
	if (send(s,pcBuf,iCnt,0) == SOCKET_ERROR)
		Fail("send");
}


void Fail(char *pszCall)
{
	printf("%s() failed: %s\n",pszCall,GetSocketError(Errno));
}

typedef struct sErr {
	int iErr;
	char *pszErr;
} SERROR;

extern SERROR sError[];

char *GetSocketError(int iError)
{
	int i;

	for (i = 0;sError[i].iErr != 0x7fff;++i) {
		if (iError == sError[i].iErr)
			break;
	}
	return sError[i].pszErr;
}

SERROR sError[] = {
  {WSABASEERR,         	"No Error"},
  {WSAEINTR,        		"Interrupted system call"},
  {WSAEBADF,        		"Bad file number"},
  {WSAEACCES,       		"Permission denied"},
  {WSAEFAULT,       		"Bad address"},
  {WSAEINVAL,       		"Invalid argument"},
  {WSAEMFILE,    	  		"Too many open files"},
  {WSAEWOULDBLOCK,     	"Operation would block"},
  {WSAEINPROGRESS,     	"Operation now in progress"},
  {WSAEALREADY,        	"Operation already in progress"},
  {WSAENOTSOCK,        	"Socket operation on non-socket"},
  {WSAEDESTADDRREQ,    	"Destination address required"},
  {WSAEMSGSIZE,        	"Message too long"},
  {WSAEPROTOTYPE,      	"Protocol wrong type for socket"},
  {WSAENOPROTOOPT,     	"Bad protocol option"},
  {WSAEPROTONOSUPPORT, 	"Protocol not supported"},
  {WSAESOCKTNOSUPPORT, 	"Socket type not supported"},
  {WSAEOPNOTSUPP,      	"Operation not supported on socket"},
  {WSAEPFNOSUPPORT,    	"Protocol family not supported"},
  {WSAEAFNOSUPPORT,    	"Address family not supported by protocol family"},
  {WSAEADDRINUSE,      	"Address already in use"},
  {WSAEADDRNOTAVAIL,   	"Can't assign requested address"},
  {WSAENETDOWN,        	"Network is down"},
  {WSAENETUNREACH,     	"Network is unreachable"},
  {WSAENETRESET,       	"Net dropped connection or reset"},
  {WSAECONNABORTED,    	"Software caused connection abort"},
  {WSAECONNRESET,      	"Connection reset by peer"},
  {WSAENOBUFS,         	"No buffer space available"},
  {WSAEISCONN,         	"Socket is already connected"},
  {WSAENOTCONN,        	"Socket is not connected"},
  {WSAESHUTDOWN,       	"Can't send after socket shutdown"},
  {WSAETOOMANYREFS,    	"Too many references, can't splice"},
  {WSAETIMEDOUT,       	"Connection timed out"},
  {WSAECONNREFUSED,    	"Connection refused"},
  {WSAELOOP,           	"Too many levels of symbolic links"},
  {WSAENAMETOOLONG,    	"File name too long"},
  {WSAEHOSTDOWN,       	"Host is down"},
  {WSAEHOSTUNREACH,    	"No Route to Host"},
  {WSAENOTEMPTY,       	"Directory not empty"},
  {WSAEPROCLIM,        	"Too many processes"},
  {WSAEUSERS,          	"Too many users"},
  {WSAEDQUOT,          	"Disc Quota Exceeded"},
  {WSAESTALE,          	"Stale NFS file handle"},
  {WSAEREMOTE,         	"Too many levels of remote in path"},
  {WSASYSNOTREADY,     	"Network SubSystem is unavailable"},
  {WSAVERNOTSUPPORTED, 	"WINSOCK DLL Version out of range"},
  {WSANOTINITIALISED,  	"Successful WSASTARTUP not yet performed"},
  {WSAHOST_NOT_FOUND,  	"Host not found"},
  {WSATRY_AGAIN,       	"Non-Authoritative Host not found"},
  {WSANO_RECOVERY,     	"Non-Recoverable errors: FORMERR, REFUSED, NOTIMP"},
  {WSANO_DATA,         	"Valid name, no data record of requested type"},
  {0x7fff,				  	"Unkown error"}
};

/* Telnet states */

#define TS_DATA	0	/* normal data */
#define TS_IAC		1
#define TS_WILL	2
#define TS_WONT	3
#define TS_DO		4
#define TS_DONT	5
#define TS_SB		6
#define TS_WSE		7
#define TS_TT		8
#define TS_3270	9

/* Telnet commands */
#define	TN_IAC	255	/* Interpret as command */
#define	TN_SE		240
#define	TN_SB		250
#define	TN_WILL	251
#define	TN_WONT	252
#define	TN_DO		253
#define	TN_DONT	254

/* Telnet subnegotiation */

#define TN_IS		0
#define TN_ARE		1
#define TN_SEND	1

/* Telnet options */
#define	TN_TRANSMIT_BINARY	0
#define	TN_ECHO					1
#define	TN_RCP					2
#define	TN_SUPPRESS_GA			3
#define	TN_NAMS					4
#define	TN_STATUS				5
#define	TN_TIMING_MARK			6
#define	TN_RCTE					7
#define	TN_NAOL					8	/* Negotiate about o/p line width */
#define	TN_NAOP					9	/* Negotiate about o/p page size */
#define	TN_NAOCRD				10	/* Negotiate about o/p CR dispos */
#define	TN_NAOHTS				11	/* Negotiate about o/p hor tab stops */
#define	TN_NAOHTD				12	/* Negotiate about o/p hor tab disp */
#define	TN_NAOFFD				13	/* Negotiate about o/p FF disp */
#define	TN_NAOVTS				14	/* Negotiate about vertical tabstops */
#define	TN_NAOVTD				15	/* Negotiate about vertical tab disp */
#define	TN_NAOLFD				16	/* Negotiate about o/p LF disp */
#define	TN_EXTEND_ASCII		17
#define	TN_LOGOUT				18
#define	TN_BM						19
#define	TN_DET					20
#define	TN_SUPDUP				21
#define	TN_SUPDUP_OUTPUT 		22
#define	TN_SEND_LOCATION 		23
#define	TN_TERMINAL_TYPE 		24
#define	TN_END_OF_RECORD 		25
#define	TN_TUID					26
#define	TN_OUTMRK				27
#define	TN_3270_REGIME			29
#define	TN_NAWS					31
#define	TN_TERMINAL_SPEED		32
#define	TN_TOGGLE_FLOW_CONTROL 33
#define	TN_LINEMODE				34
#define	TN_EXOPL					255
#define TN_TN3270E				40
#define TN_IMPLEMENTED			40	/* highest number implemented */

/* implemented options */

#define	TNI_TRANSMIT_BINARY	0
#define	TNI_ECHO					1
#define	TNI_SUPPRESS_GA		2
#define	TNI_TERMINAL_TYPE		3
#define	TNI_END_OF_RECORD		4
#define	TNI_3270_REGIME		5
#define	TNI_NAWS					6
#define	TNI_TERMINAL_SPEED	7
#define	TNI_TOGGLE_FLOW_CONTROL	8
#define	TNI_LINEMODE			9
#define TNI_TN3270E				10
#define TNI_SEND_LOCATION		11
#define	NOPTIONS					12


char arcTelnetOptions[TN_IMPLEMENTED + 1] =
{	/* implemented options */
	TNI_TRANSMIT_BINARY,/* TN_TRANSMIT_BINARY */
	TNI_ECHO,			/* TN_ECHO */
	NOPTIONS,			/* TN_RCP */
	TNI_SUPPRESS_GA,	/* TN_SUPPRESS_GA */
	NOPTIONS,			/* TN_NAMS */
	NOPTIONS,			/* TN_STATUS */
	NOPTIONS,			/* TN_TIMING_MARK */
	NOPTIONS,			/* TN_RCTE */
	NOPTIONS,			/* TN_NAOL */
	NOPTIONS,			/* TN_NAOP */
	NOPTIONS,			/* TN_NAOCRD */
	NOPTIONS,			/* TN_NAOHTS */
	NOPTIONS,			/* TN_NAOHTD */
	NOPTIONS,			/* TN_NAOFFD */
	NOPTIONS,			/* TN_NAOVTS */
	NOPTIONS,			/* TN_NAOVTD */
	NOPTIONS,			/* TN_NAOLFD */
	NOPTIONS,			/* TN_EXTEND_ASCII */
	NOPTIONS,			/* TN_LOGOUT */
	NOPTIONS,			/* TN_BM */
	NOPTIONS,			/* TN_DET */
	NOPTIONS,			/* TN_SUPDUP */
	NOPTIONS,			/* TN_SUPDUP_OUTPUT */
	NOPTIONS,			/* TN_SEND_LOCATION */
	TNI_TERMINAL_TYPE,/* TN_TERMINAL_TYPE */
	TNI_END_OF_RECORD,/* TN_END_OF_RECORD */
	NOPTIONS,			/* TN_TUID */
	NOPTIONS,			/* TN_OUTMRK */
	NOPTIONS,			/* ????????? */
	TNI_3270_REGIME,	/* TN_3270_REGIME */
};

char arcLocal[NOPTIONS];
char arcRemote[NOPTIONS];
char cDontSent;		//already sent dont for specified option, don't send again
char cWontSent;      //same as above

#define TO_NOSET  0x80	/* WONT a DO request and DONT a WILL request */
#define TO_SENT	0x40	/* option already sent */
#define TO_TTRCVD 0x20	/* Terminal type has been received */
#define TO_INIT	0x10	/* Initialize negotiation on invokation of client */
#define TO_SET	   0x01	/* DO or WILL is set */

static char arcSubEnd[] = {TN_IAC, TN_SE};
static char arcSubBegin[] = {TN_IAC, TN_SB, TN_TERMINAL_TYPE, TN_IS};
static char arcTermType[] = "dumb";

/*
Initiate telnet. Our negotionation preferences may be set here.

This function currently does nothing.
*/
void InitTelnet(void)
{
//	arcLocal[TNI_TERMINAL_TYPE] |= TO_NOSET;
}

/*
Strip and handle Telnet options. The data in the buffer is compressed to
contain only ordinary (non-telnet) data bytes. The length of the compressed
data is returned.

Arguments:
	pcBuf - Pointer to buffer containing telnet data to handle
	iCnt - Count of octets to handle

Returns:
	The amount of normal data bytes remaining in the buffer.
*/
int TeloptStrip(char *pcBuf, int iCnt)
{
	int iRetCnt;
	char *pcReadPos;
	char *pcWritePos;
	u_char bCh;

	/* Optimization for very common special case -- no TNOPT */
	if (iTelnetState == TS_DATA && memchr(pcBuf, TN_IAC, iCnt) == NULL)
	{
		return iCnt;
	}
	pcWritePos = pcReadPos = pcBuf;
	iRetCnt = 0;

	while (iCnt--)
	{
		bCh = *pcReadPos++;
		switch(iTelnetState)
		{
		case TS_DATA:
			if (bCh == TN_IAC)
			{
				iTelnetState = TS_IAC;
			}
			else
			{
				*pcWritePos++ = bCh;
				++iRetCnt;
			}
			break;
		case TS_IAC:
			switch (bCh)
			{
			case TN_WILL:
				iTelnetState = TS_WILL;
				break;
			case TN_WONT:
				iTelnetState = TS_WONT;
				break;
			case TN_DO:
				iTelnetState = TS_DO;
				break;
			case TN_DONT:
				iTelnetState = TS_DONT;
				break;
			case TN_IAC:
				*pcWritePos++ = bCh;
				++iRetCnt;
				iTelnetState = TS_DATA;
				break;
			case TN_SB:
				iTelnetState = TS_SB;
				break;
			default:
				iTelnetState = TS_DATA;
				break;
			}
			break;
		case TS_WILL:
			WillOpt(bCh);
			iTelnetState = TS_DATA;
			break;
		case TS_WONT:
			WontOpt(bCh);
			iTelnetState = TS_DATA;
			break;
		case TS_DO:
			DoOpt(bCh);
			iTelnetState = TS_DATA;
			break;
		case TS_DONT:
			DontOpt(bCh);
			iTelnetState = TS_DATA;
			break;
		case TS_SB:
			if (bCh == TN_IAC)
				iTelnetState = TS_IAC;
			else if (bCh == TN_3270_REGIME)
				iTelnetState = TS_3270;
			else if (bCh == TN_TERMINAL_TYPE)
				iTelnetState = TS_TT;
			else
				iTelnetState = TS_WSE;
			break;
		case TS_3270:
			if (bCh == TN_IAC)
				iTelnetState = TS_IAC;
			break;
		case TS_WSE:		/* ignore up to next IAC */
			if (bCh == TN_IAC)
				iTelnetState = TS_IAC;
			break;
		case TS_TT:
			if (bCh == TN_SEND)
			{
				char *p;
				int len;
				char nbuf[50];

				memcpy(nbuf, arcSubBegin, sizeof(arcSubBegin));
				strcpy(nbuf + (len = sizeof(arcSubBegin)),
						p = arcTermType);
				memcpy(nbuf + (len += strlen(p)), arcSubEnd,
								sizeof(arcSubEnd));
				SendData(nbuf, len + sizeof(arcSubEnd));
			}
			else if (bCh == TN_IS)
				arcRemote[TNI_TERMINAL_TYPE] |= TO_TTRCVD;
			iTelnetState = TS_WSE;
			break;
		}
	}
	return iRetCnt;
}

/*
Handle a 'will' sent by the other party by replying on it based on the
preferences in the arcRemote array.

Argument:
	cOption - The option to be negotiated.
*/
void WillOpt(char cOption)
{
	char cAck = TN_DONT;
	int iOptionIndex = arcTelnetOptions[cOption];
	char *pcOption;

	if (cOption <= TN_IMPLEMENTED && iOptionIndex < NOPTIONS)
	{
		pcOption = arcRemote + iOptionIndex;
		if (*pcOption & (TO_SET | TO_SENT))
		{
			/* Already set or sent, ignore to prevent loop */
			*pcOption = (char)((*pcOption & ~TO_SENT) | TO_SET);
			return;
		}
		if (!(*pcOption & TO_NOSET))
		{
			*pcOption |= TO_SET;
			cAck = TN_DO;
		}
	}
	else if (cOption == cDontSent)
		return;		/* have already sent a DONT for this one */
	else
		cDontSent = cOption;
	SendOpt(cAck, cOption);
}

/*
Handle a 'wont' sent by the other party by replying on it based on the
preferences in the arcRemote array.

Argument:
	cOption - The option to be negotiated.
*/
void WontOpt(char cOption)
{
	int iOptionIndex = arcTelnetOptions[cOption];
	char *pcOption;
	char oldopt;

	if (cOption <= TN_IMPLEMENTED && iOptionIndex < NOPTIONS)
	{
		oldopt = *(pcOption = arcRemote + iOptionIndex);
		*pcOption &= ~(TO_SET | TO_SENT);
		if ((oldopt & (TO_SENT | TO_SET)) == TO_SENT)
		{
			/* Already clear, ignore to prevent loop */
			return;
		}
	}
	else if (cOption == cDontSent)
		return;		/* have already sent a DONT for this one */
	else
		cDontSent = cOption;
	SendOpt(TN_DONT, cOption);	/* Must always accept */
}

/*
Handle a 'do' sent by the other party by replying on it based on the
preferences in the arcLocal array.

Argument:
	cOption - The option to be negotiated.
*/
void DoOpt(char cOption)
{
	char cAck = TN_WONT;
	int iOptionIndex = arcTelnetOptions[cOption];
	char *pcOption;

	if (cOption <= TN_IMPLEMENTED && iOptionIndex < NOPTIONS)
	{
		pcOption = arcLocal + iOptionIndex;
		if (*pcOption & (TO_SET | TO_SENT))
		{
			/* Already set or sent, ignore to prevent loop */
			*pcOption = (char)((*pcOption & ~TO_SENT) | TO_SET);
			return;
		}
		if (!(*pcOption & TO_NOSET))
		{
			*pcOption |= TO_SET;
			cAck = TN_WILL;
		}
	}
	else if (cOption == cWontSent)
		return;		/* have already sent a WONT for this one */
	else
		cWontSent = cOption;
	SendOpt(cAck, cOption);
}

/*
Handle a 'dont' sent by the other party by replying on it based on the
preferences in the arcLocal array.

Argument:
	cOption - The option to be negotiated.
*/
void DontOpt(char cOption)
{
	int iOptionIndex = arcTelnetOptions[cOption];
	char *pcOption, oldopt;

	if (cOption <= TN_IMPLEMENTED && iOptionIndex < NOPTIONS)
	{
		oldopt = *(pcOption = arcLocal + iOptionIndex);
		*pcOption &= ~(TO_SET | TO_SENT);
		if ((oldopt & (TO_SENT | TO_SET)) == TO_SENT)
		{
			/* Already clear, ignore to prevent loop */
			return;
		}
	}
	else if (cOption == cWontSent)
		return;		/* have already sent a WONT for this one */
	else
		cWontSent = cOption;
	SendOpt(TN_WONT, cOption);
}

/*
Send a telnet command to the server for the specified option.

arguments:
	cCommand - The Telnet command (will, wont, etc)
	cOption - The telnet option (eg echo)
*/
void SendOpt(char cCommand, char cOption)
{
	char arcSend[3];
	int iOptionIndex;

	if ((iOptionIndex = arcTelnetOptions[cOption]) < NOPTIONS)
	{
		//record that we sent this command to prevent loop.
		*((((u_char)cCommand == TN_WILL || (u_char)cCommand == TN_WONT) ? arcLocal : arcRemote)
		 +	iOptionIndex) |= TO_SENT;
	}
	arcSend[0] = TN_IAC;
	arcSend[1] = cCommand;
	arcSend[2] = cOption;
	SendData(arcSend, 3);
}

