//  $Id: UDPCHAT.C 1.9 2003/07/23 17:09:20 peterb Exp $
//
//  *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//  Copyright (c) 1999-2003 Datalight, Inc.
//  All Rights Reserved
//
//  Datalight, Inc. is a Washington corporation located at:
//
//  21520 30th Dr SE, Suite 110,          Tel: +1.425.951.8086
//  Bothell, WA  98021                    Fax: +1.425.951.8094
//  USA                                   Http://www.datalight.com
//
//  This software, including without limitation all source code and
//  documentation, is the confidential, trade secret property of
//  Datalight, Inc., and is protected under the copyright laws of
//  the United States and other jurisdictions.  Portions of the
//  software may also be subject to one or more the following US
//  patents: US#5860082, US#6260156.
//
//  In addition to civil penalties for infringement of copyright
//  under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//  penalties for violation of (a) the restrictions on circumvention
//  of copyright protection systems found in 17 U.S.C. 1201 and
//  (b) the protections for the integrity of copyright management
//  information found in 17 U.S.C. 1202.
//
//  U.S. Government Restricted Rights.  Use, duplication,
//  reproduction, or transfer of this commercial product and
//  accompanying documentation is restricted in accordance with
//  FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//  IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//  PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//  AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//  INC. AND THE LICENSEE ("BINDING AGREEMENT").  YOUR RIGHT, IF ANY,
//  TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//  TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//  USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//  TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT.  CONTACT
//  DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//  THIS SOFTWARE IN ERROR.
//
/*
	$DOCFILE:UDPCHAT.C

	Copyright (C) 1999-2001 by Datalight, Inc.
	All Rights Reserved.

	A program that enables users to chat using udp.

	'Chat' means to to talk and to listen. When using UDP, it means that for the
	talk size we just send a broadcast message on the lan and for the listen side
	we start a UDP server.

	This program and its functions are non-reentrant.

	$DOCHISTORY
	11/07/2001 jmb removed call to GetPeerAddress and added paramter to
	               ReadSocket to get the IP address of sender.
	04/30/2001 jmb displayed Datalight copyright message.
	04/17/2001 jmb fixed problem with received messages not being displayed
	               when not displaying IP address of sender with message by
	               adding received message to print statement.
	06/29/1999 jmb added copyright header
*/

#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <stdlib.h>
#include <conio.h>
#include <process.h>
#include "compiler.h"
#include "capi.h"
#include "udpchat.pt"

#define CHAT_PORT 33
#define BUF_SIZE 200
#define NAME_LEN	20	//my name

static char rgcKeyBuf[BUF_SIZE];//for editing
static int iKeyCount = 0; //number of characters in keybuf
static NET_ADDR sNetAddr; //general use
static NET_ADDR sNetAddrRmt; //general use

/*
Sits in a loop and check for two things:
1. If any data was received from the network
	If data was received, it is displayed on screen
2. If the user has entered any data
	If the character entered was a newline, the previously entered data is
	send, else the character is stored in a buffer.
*/
void main(void)
{
	char rgcMyNameBuf[NAME_LEN];//buffer containing my (arbritary) name
	char *pMyName;		//pointer to it
	int iMaxInput;		//maximum no of bytes before we must send (to keep all in buffer)
	char rgcBuf[BUF_SIZE];//general buffer, used for sending and receiving
	int iRcvSock;		//socket for receiving packets
	int iSendSock;		//socket for sending packets
	int iLength;		//Length of string read or written
	char cCh;         //Character read
	char cMoreInfo;	//Whether or not to display additional information

	printf("Sockets UDP Chat Client\n(dt)");
	printf("Copyright (C) 1999-2001 Datalight, Inc.\nAll Rights Reserved\n"
	       "Portions copyright (c) GPvNO 2003\n\n");

	//get user name and set variables accordingly
	printf("Enter your name (send with all your messages):");
	rgcMyNameBuf[0] = NAME_LEN - 2;
	pMyName = cgets(rgcMyNameBuf);
	iMaxInput = BUF_SIZE - rgcMyNameBuf[1] - 10;

	//get more info option
	printf("\nDo you want to see the ip addresses of senders? [Y/N]");
	rgcBuf[0] = 2;
	cgets(rgcBuf);
	cMoreInfo = (rgcBuf[2] == 'Y' ? 1 : 0);

	printf("\nPress Alt-X to exit\n");

	//start to listen for datagrams
	iSendSock = GetClientSock();
	iRcvSock = StartListen(CHAT_PORT);
	if (iRcvSock == 0 || iSendSock == 0)
		return;
	//tell the world im on the air
	iLength = sprintf(rgcBuf, "%s came on the air", pMyName);
	if (WriteSocket(iSendSock, rgcBuf, iLength, NET_FLG_BROADCAST) < 0)//error
		Aprintf("Error on Sending %d bytes: %s\n",iLength, Err(iNetErrNo));

	while (1)
	{
		//server part - see if we received data
		iLength = ReadSocket(iRcvSock, rgcBuf, BUF_SIZE, &sNetAddr, 0);
		if (iNetErrNo != 0 && iNetErrNo != ERR_WOULD_BLOCK)
			Aprintf("Error on Netread: %s\n", Err(iNetErrNo));
		if (iLength > 0)
		{
			rgcBuf[iLength] = 0;
			//no color now l = sNetAddr.dwRemoteHost & 0xff;//some color identifing host
			if (cMoreInfo)//give the senders ip address as well
			{
				Aprintf("%s (from %s)\n", rgcBuf, WriteName(&sNetAddr));
			}
			else
				Aprintf("%s\n", rgcBuf);
		}
		if (kbhit())
		{
			switch(cCh = getche())
			{
			case 0:				/* function key */
				if (getch() == 45)//Alt-X  exit
				{
					ReleaseSocket(iRcvSock);
					ReleaseSocket(iSendSock);
					return;
				}
				break;
			default:
				if (iKeyCount < iMaxInput)
				{
					rgcKeyBuf[iKeyCount++] = cCh;
					break;
				}
				//fall thru
			case '\r':
			//case '\n':
				putch('\n');
				if (iKeyCount == 0)//nothing to do
					break;
				//everyone should identify himself
				rgcKeyBuf[iKeyCount] = 0;
				iLength = sprintf(rgcBuf, "%s :%s", pMyName, rgcKeyBuf);
				//broadcast data
				if (WriteSocket(iSendSock, rgcBuf, iLength, NET_FLG_BROADCAST) < 0)//error
					Aprintf("Error on NetWrite %d bytes: %s\n", iLength, Err(iNetErrNo));
				iKeyCount = 0;
				break;
			}
		}
	}
}

/*
Get a socket descriptor that can be used to send UDP packets

Returns:
	A descriptor of the socket, or 0 if an error occured. (Returns immediately)
*/
static int GetClientSock(void)
{
	int iSock;

	memset(&sNetAddr, 0, sizeof(NET_ADDR));
	sNetAddr.wRemotePort = CHAT_PORT;

	printf("%d: sNetAddr.wLocalPort = %d\n", __LINE__, sNetAddr.wLocalPort);
	printf("%d: sNetAddr.wRemotePort = %d\n", __LINE__, sNetAddr.wRemotePort);
	

	if ((iSock = GetSocket()) < 0)
		Aprintf("Error on GetSocket(): %s\n",Err(iNetErrNo));
	else if (SetSocketOption(iSock, 0, NET_OPT_NON_BLOCKING, 1, 1) < 0)
		Aprintf("Error on setOpt(): %s\n",Err(iNetErrNo));
	else if (ConnectSocket(iSock, DATA_GRAM, &sNetAddr) < 0)
		Aprintf("Error on net_connect: %s\n",Err(iNetErrNo));
	else
	{
		printf("%d: sNetAddr.wLocalPort = %d\n",  __LINE__, sNetAddr.wLocalPort);
		printf("%d: sNetAddr.wRemotePort = %d\n", __LINE__, sNetAddr.wRemotePort);
		Aprintf("Client sock successfully created\n");
		return iSock;
	}

	return 0;
}
/*
Start a UDP server on specified port.

Returns:
	A descriptor of the server socket, or 0 if an error occured.
	(Returns immediately)

*/
int StartListen(int iPort)
{
	int iServSock;	//the new socket

//	memset(&sNetAddr, 0, sizeof(NET_ADDR));
	sNetAddr.wLocalPort = iPort;

	if ((iServSock = GetSocket()) < 0)
		Aprintf("Error on serv GetSocket(): %s\n",Err(iNetErrNo));
	else if (SetSocketOption(iServSock, 0, NET_OPT_NON_BLOCKING, 1, 1) < 0)
		Aprintf("Error on serv setOpt(): %s\n", Err(iNetErrNo));
	else if (ListenSocket(iServSock, DATA_GRAM, &sNetAddr) < 0)
		Aprintf("Error on serv net_listen: %s\n", Err(iNetErrNo));
	else
	{
		Aprintf("Server sock successfully created\n");
		return iServSock;
	}
	return 0;
}

/*
Creates a human understandible string from a Sockets error code

Argument:
	uErrCode - The sockets error code

Returns:
	A pointer to the (static) string representation of the error
*/
char *Err(unsigned uErrCode)
{
	static char rgcUnk[30];
	static char *rgszErrs[] =
	{
		"NoErr",
		"InUse",
		"DOSErr",
		"NoMem",
		"NotNetconn",
		"IllegalOp",
		"BadPkt",
		"NoHost",
		"CantOpen",
		"NetUnreachable",
		"HostUnreachable",
		"ProtUnreachable",
		"PortUnreachable",
		"TimeOut",
		"HostUnknown",
		"NoServers",
		"ServerErr",
		"BadFormat",
		"BadArg",
		"EOF",
		"Reset",
		"WouldBlock",
		"UnBound",
		"NoDesc",
		"BadSysCall",
		"CantBroadcast",
		"NotEstab",
		"ReEntry",
	};

	if (uErrCode == ERR_API_NOT_LOADED)
		return "Sockets API not loaded";
	if ((uErrCode & 0xff) > ERR_RE_ENTRY)
	{
		sprintf(rgcUnk,"Unknown error 0x%04X",uErrCode);
		return rgcUnk;
	}
	return rgszErrs[uErrCode & 0xff];
}


/*
Show data on screen.

For now we just use vprintf to print the data and reprint any stuff that was
being edited.

Later we may put each person's data in a seperate block on screen, or whatever
seems fancy.

Arguments:
	Format and eclipse - Exactly the same as for prinf().
*/
void Aprintf (char *pFormat, ...)
{
	va_list pArgs;

	if (iKeyCount)//data been edited, start on newline
		printf("\n");

	va_start(pArgs, pFormat);
	vprintf(pFormat, pArgs);
	va_end(pArgs);

	//print the old edited stuff (if any)
	if (iKeyCount)
	{
		rgcKeyBuf[iKeyCount] = 0;
		printf("%s", rgcKeyBuf);
	}
}

/*
Create a string representation of the IP address and port, in the form
	a.b.c.d:port, eg 196.10.180.3:1400.
Max length is 22 bytes.

Arguments:
	psAddr - pointer to NET_ADDR structure containing address of host.

Returns:
	Pointer to (static) array containing null-terminated string.
*/
static char *WriteName(NET_ADDR *psAddr)
{
	static char rgcName[22];

	sprintf(rgcName, "%u.%u.%u.%u:%u",
		((BYTE *)&psAddr->dwRemoteHost)[0],
		((BYTE *)&psAddr->dwRemoteHost)[1],
		((BYTE *)&psAddr->dwRemoteHost)[2],
		((BYTE *)&psAddr->dwRemoteHost)[3],
		psAddr->wRemotePort
	);
	return rgcName;
}

