//
//   $Id: Sockdiag.c 1.7 2003/08/13 09:00:58 johnb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*
	$DOCFILE:SOCKDIAG.C

	Copyright (C) 1999-2001 Datalight, Inc.
	All Rights Reserved

	Diagnostic testing Sockets/DOS using BSD Sockets

	$DOCHISTORY:
	2003-07-21	gvn explicit test for ioctlsocket(FIONREAD) setting 32 bits
	2003-07-03	gvn re-fixed WinSock inconsistency in the "Conflicting type and protocol" test
	2003-01-07	gvn added tests for database functions, some fixes
	2003-01-02	gvn added tests for select, get/setsockopt
	2002-12-10	gvn first version ported from CAPIDIAG

	Compile and link with SOCKET.C, CAPI.C and _CAPI.C
	Use TABS=3 spaces for properly formatted source

	SOCKDIAG requires a remote server with both a UDP and TCP echo server
	for a complete test. NT server, Unix or Inet can provide this. It can
	also be run without such a server as a stand-alone test using only the
	internal loopback capabilities of Sockets.
*/


#ifdef __DJGPP__
#include <dpmi.h>
#include <go32.h>
#endif

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <conio.h>
#include <process.h>
#include <stdarg.h>

#define _POSIX_SOURCE
#include <time.h>
#undef  _POSIX_SOURCE

#include <dos.h>
#include <ctype.h>
#include <signal.h>
#include <errno.h>

#ifdef _Windows
#include <winsock2.h>
#define GET_MILLIS clock()
#define FACTOR 1000.0
#else
#include "compiler.h"
#include "socket.h"
#define GET_MILLIS (GET_TICKS * 55L)
#define FACTOR 998.7571
#endif

#ifndef __WIN_TYPES
 #define __WIN_TYPES
 typedef unsigned char BYTE;
 typedef unsigned short WORD;
 typedef unsigned long DWORD;
#endif

//
// prototypes
//

#include "sockdiag.pt"

#ifdef NULL_POINTER_TEST
char arcNullData[30];
#endif

SOCKET soClient, soServer = INVALID_SOCKET, soListen = INVALID_SOCKET;
struct sockaddr_in sClientAddr, sServerAddr;

int iFailAction;		// action on failure
int iFail;				// fail/pass
int iVerbose;			// display tests
WORD wTimeOut = 10;	// timeout value
WORD wTimeOutHalf = 5;
int iFailures;
int iServerFailures;
int iNumberTests;
int iWriteSize, iReadSize, iReadOffset, iWriteOffset;
int iBlockError,iBlocking;
int iErrorAddress;
DWORD dwBufsWritten, dwBufsRead;
long lTime, lTarget, lTimeRead;
BYTE bWriting;
BYTE bReset;		// causes server to reset connection
BYTE bDiagnostic;	// Between Start() and End()
WORD wRead,wWrite;
int iBulkTime = 1000;

char szData[] = "0123456789abcdefghijklmnopqrstuvwxyzABC";
#define BUFLEN		1024
char arcReadBuf[BUFLEN];
char rgcWriteBuf[BUFLEN];
DWORD dwMyIPAddress;
DWORD dwServerIPAddress;
BYTE bDebugServer;		// debug the server
BYTE bServerInUse;
volatile int iControlBreak;
char szDiagnostic[400];	// Currently running diagnostic
#define WRITE_AHEAD	3	// # of blocks to write ahead in bulk transfer tests
WORD wLocalPort = 7,wRemotePort = 7,wPort;
char *arszProt[] = {"tcp","udp",0};
char *arszServ[] = {"discard","sink","null"};

void ControlBreak(int iSignal)
{
	iControlBreak = iSignal;
}

int main(int argc, char *argv[])
{
	int iProt,iServ,iLen,i;
	char szNameBuf[80];
	char *pszArg,**ppsz;
	SOCKET s;
	SOCKADDR_IN sRemoteAddress;
	SOCKADDR_IN sIn;
	int iAddrSize = sizeof(SOCKADDR),iNum;
	struct hostent *psHost;
	struct protoent *psProtoEnt;
	struct servent *psServEnt;
	DWORD dwRemoteIPAddress = 0;

	WSADATA 	winsockdata;

	printf("Datalight %d bit Sockets Diagnostic version 1.04 (%s)\n",
																sizeof(int) * 8,__DATE__);
	printf("Copyright (C) 1999-2003 Datalight, Inc.\nAll Rights Reserved\n"
			 "Portions copyright (c) GPvNO 2003\n\n");
	while (argc > 1 && argv[1][0] == '/')
	{
		pszArg = argv[1] + ((argv[1][2] == '=') ? 3 : 2);
		switch (tolower(argv[1][1]))
		{
			case 'b':	// bulk test time in seconds
				if ((iBulkTime = atoi(pszArg) * 1000) == 0)
					iBulkTime = 1000;
				break;
			case 'd':
#if defined(SOCK_DEBUG) && !defined _Windows
				if (StartDebug(pszArg) < 0)
					printf("Unable to open %s\n",pszArg);
				break;
#endif
			case 's':
				bDebugServer = 1;
				break;
			default:
				printf(	"SOCKDIAG [/d][/l=<level>][/v=<verbose>] [<server>]\n"
							"/b=<bulk_test_seconds>\n"
#if defined(SOCK_DEBUG) && !defined _Windows
							"/d=<debug_file>\n"
#endif
							"/f=<fail_action>\n"
							"   0 print error and continue\n"
							"   1 Print error and wait for Esc\n"
							"   2 Print error and exit\n"
							"/l=<local_echo_port> (default 7)\n"
							"/r=<remote_echo_port> (default 7)\n"
							"/s - debug Server\n"
							"/t=<timeout_seconds>\n"
							"/v=<verbosity> add values below\n"
							"   0 only print errors\n"
							"   1 print names of all tests\n"
							"   2 print additional information on errors\n"
							"   4 print summary of events\n"
							"   8 print kernel configuration\n");
				exit(0);
			case 'f':
				iFailAction = atoi(pszArg);
				break;
			case 'l':
				wLocalPort = (WORD)atoi(pszArg);
				break;
			case 'r':
				wRemotePort = (WORD)atoi(pszArg);
				break;
			case 't':
				wTimeOut = (WORD)atoi(pszArg);
				if ((wTimeOutHalf = (WORD)(wTimeOut / 2)) == 0)
					wTimeOutHalf = 1;
				break;
			case 'v':
				iVerbose = atoi(pszArg);
				break;
		}
		--argc;
		++argv;
	}
#ifdef NULL_POINTER_TEST
	memcpy(arcNullData,(char *)0,sizeof(arcNullData));
#endif

	Start("WSAStartup()");
	if ((i = WSAStartup(0x0101,&winsockdata)) != 0) {
		WSACleanup();
		printf("WSAStartup failed: %s\n",GetErrorString(i));
		exit(1);
	}
	End();
	printf("Sockets API v %u.%u\n%s\nSockets available: %d\n",
		HIBYTE(winsockdata.wHighVersion),
		LOBYTE(winsockdata.wHighVersion),
		winsockdata.szDescription,winsockdata.iMaxSockets);

	iLen = sizeof(szNameBuf);
	szNameBuf[0] = 0;

	Start("gethostname()");
	if (gethostname(szNameBuf, iLen) == SOCKET_ERROR)
	{
		if (WSAGetLastError() == WSAENETDOWN)
		{
			printf("Network down\n");
			WSACleanup();
			exit(1);
		}
		Fail(0);
	}
	End();

	printf("Socket Diagnostic on '%s'\n", szNameBuf);
	signal(SIGINT, ControlBreak);

	for (iServ = 0;iServ < 3;++iServ)
	{
		for (iProt = 0;iProt < 3;++iProt)
		{
			Start("getservbyname(%s,%s)",arszServ[iServ],arszProt[iProt] ?
																arszProt[iProt] : "0");
			if ((psServEnt = getservbyname(arszServ[iServ],arszProt[iProt])) == 0)
				Fail(0);
			else if (strcmp(psServEnt->s_name,"discard") != 0)
				Fail("&s_name: %s expected: discard",psServEnt->s_name);
			else if (ntohs(psServEnt->s_port) != 9)
				Fail("&s_port: %d, expected: 9",ntohs(psServEnt->s_port));
			else if (strcmp(psServEnt->s_proto,"tcp") != 0 &&
												strcmp(psServEnt->s_proto,"udp") != 0)
				Fail("&s_proto: %s expected: tcp or udp",psServEnt->s_proto);
			else for (ppsz = psServEnt->s_aliases,iNum = 0;*ppsz;++iNum,++ppsz)
			{
				if (strcmp(*ppsz,"null") != 0 &&
								strcmp(*ppsz,"sink") != 0)
				{
					Fail("&s_alias: %s",*ppsz);
					break;
				}
			}
			if (!*ppsz && iNum != 2)
				Fail("&Aliases: %d expected: 2",iNum);
			End();
		}
	}

	for (iProt = 0;iProt < 3;++iProt)
	{
		Start("getservbyport(htons(9),%s)",arszProt[iProt] ?
																arszProt[iProt] : "0");
		if ((psServEnt = getservbyport(htons(9),arszProt[iProt])) == 0)
			Fail(0);
		else if (strcmp(psServEnt->s_name,"discard") != 0)
			Fail("&s_name: %s expected: discard",psServEnt->s_name);
		else if (ntohs(psServEnt->s_port) != 9)
			Fail("&s_port: %d, expected: 9",ntohs(psServEnt->s_port));
		else if (strcmp(psServEnt->s_proto,"tcp") != 0 &&
											strcmp(psServEnt->s_proto,"udp") != 0)
			Fail("&s_proto: %s expected: tcp or udp",psServEnt->s_proto);
		else for (ppsz = psServEnt->s_aliases,iNum = 0;*ppsz;++iNum,++ppsz)
		{
			if (strcmp(*ppsz,"null") != 0 &&
							strcmp(*ppsz,"sink") != 0)
			{
				Fail("&s_alias: %s",*ppsz);
				break;
			}
		}
		if (!*ppsz && iNum != 2)
			Fail("&Aliases: %d expected: 2",iNum);
		End();
	}

	Start("getprotobyname(\"tcp\")");
	if ((psProtoEnt = getprotobyname("tcp")) == 0)
		Fail(0);
	else if (psProtoEnt->p_proto != 6)
		Fail("&p_proto: %d, expected: 6",psProtoEnt->p_proto);
	else if (strcmp(psProtoEnt->p_name,"tcp") != 0)
		Fail("&p_name: %s expected: tcp",psProtoEnt->p_name);
	else for (ppsz = psProtoEnt->p_aliases,iNum = 0;*ppsz;++ppsz,++iNum)
	{
		if (strcmp(*ppsz,"TCP") != 0)
		{
			Fail("&p_alias %s",*ppsz);
			break;
		}
	}
	if (!*ppsz && iNum != 1)
		Fail("&Aliases: %d expected: 1",iNum);
	End();

	Start("getprotobynumber(6)");
	if ((psProtoEnt = getprotobynumber(6)) == 0)
		Fail(0);
	else if (psProtoEnt->p_proto != 6)
		Fail("&p_proto: %d, expected: 6",psProtoEnt->p_proto);
	else if (strcmp(psProtoEnt->p_name,"tcp") != 0)
		Fail("&p_name: %s expected: tcp",psProtoEnt->p_name);
	else for (ppsz = psProtoEnt->p_aliases,iNum = 0;*ppsz;++ppsz,++iNum)
	{
		if (strcmp(*ppsz,"TCP") != 0)
		{
			Fail("&p_alias %s",*ppsz);
			break;
		}
	}
	if (!*ppsz && iNum != 1)
		Fail("&Aliases: %d expected: 1",iNum);
	End();

	if (argc > 1)
	{
		Start("inet_addr(\"%s\")",argv[1]);
		if ((dwRemoteIPAddress = inet_addr(argv[1])) == INADDR_NONE)
		{
			End();
			Start("gethostbyname(\"%s\")",argv[1]);
			if	((psHost = gethostbyname(argv[1])) == 0)
				Fail(0);
			else
				dwRemoteIPAddress = *(DWORD *)(psHost->h_addr_list[0]);
		}
		End();
	}
	// determine local IP address
	Start("socket(SOCK_DGRAM)");
	if ((s = socket(PF_INET,SOCK_DGRAM,0)) == INVALID_SOCKET)
		Fail(0);
	End();
	// Connect to arbitrary port and address (NOT loopback)
	sRemoteAddress.sin_family = AF_INET;
	sRemoteAddress.sin_port	 = htons(IPPORT_ECHO);
	sRemoteAddress.sin_addr.s_addr = inet_addr("128.127.50.1");
	// Get local address
	Start("connect(SOCK_DGRAM)");
	if (connect(s,(PSOCKADDR)&sRemoteAddress,sizeof(SOCKADDR)) != SOCKET_ERROR)
	{
		End();
		Start("getsockname()");
		if (getsockname(s,(PSOCKADDR)&sIn,&iAddrSize) != SOCKET_ERROR)
		{
			if ((dwMyIPAddress = sIn.sin_addr.s_addr) == 0)
				Fail("&returned 0.0.0.0");
		}
		else
			Fail(0);
	}
	else
		Fail(0);
	End();
	Start("gethostbyname(gethostname())");
	if	((psHost = gethostbyname(szNameBuf)) == 0)
		Fail(0);
	else {
		dwServerIPAddress = *(DWORD *)(psHost->h_addr_list[0]);
		if (dwServerIPAddress != dwMyIPAddress)
			Fail("&%s",IPAddressString(dwServerIPAddress));
	}
	End();
	Start("closesocket()");
	if (closesocket(s) == SOCKET_ERROR)
		Fail(0);
	End();

	for (i = 0;psHost->h_addr_list[i];i++)
	{
		dwMyIPAddress = *(DWORD *)(psHost->h_addr_list[i]);
		printf("Local IP address:  %s\n",IPAddressString(dwMyIPAddress));
		dwServerIPAddress = dwMyIPAddress;

		wPort = wLocalPort;
		RunDiagnostics();
		if (dwRemoteIPAddress)
		{
			dwServerIPAddress = dwRemoteIPAddress;
			wPort = wRemotePort;
			RunDiagnostics();
		}
	}
	printf("Number of tests: %d, Failures: %d\n", iNumberTests, iFailures);
	WSACleanup();
	return 0;
}

void RunDiagnostics(void)
{
	char szAddress[40];
	DWORD dwAddress;

	printf("Remote IP address: %s\n", IPAddressString(dwServerIPAddress));
	sClientAddr.sin_addr.s_addr = dwServerIPAddress;
	sprintf(szAddress,"%s",IPAddressString(dwServerIPAddress));
	Start("inet_addr(\"%s\")",szAddress);
	if ((dwAddress = inet_addr(szAddress)) == 0L)
		Fail(0);
	else if (dwAddress != dwServerIPAddress)
		Fail("&returned %s", IPAddressString(dwAddress));
	End();
	// flush keyboard buffer
	while (kbhit())
		getch();

	TestSocket(SOCK_STREAM);
	TestSocket(SOCK_DGRAM);
}

void Exit(int iCode)
{
	closesocket(soClient);
	closesocket(soServer);
	closesocket(soListen);
	exit(iCode);
}

int Getch(void)
{
	int iKey;
	char *pszArgvl[2];	//used in spawning COMSPEC

	iKey = getch();
	if (iKey == 0)
		/* function key */
		iKey = 1000 + getch();
	switch (iKey)
	{
		case 1045:		/* alt X */
			printf("User exit\n");
			Exit(2);
			break;
		case 1035:	// Alt H
			printf( "\n"
					"Escape        Abort test\n"
					"Alt-E         Execute command.com\n"
					"Alt-H         Help\n"
					"Alt-X         eXit\n"
					"Any other key Display current test\n");
			break;
		case 1018:	// alt e
			if ((pszArgvl[0] = getenv("COMSPEC")) == 0)
				pszArgvl[0] = "\\COMMAND.COM";
			pszArgvl[1] = 0;
			if (spawnvp(P_WAIT, pszArgvl[0], pszArgvl) < 0)
				perror("exec error");
			break;
	}
	return iKey;
}

// general test for select()

void TestSelect(SOCKET s,fd_set *pfdsRead,fd_set *pfdsWrite,fd_set *pfdsExcept,
		struct timeval *psTimeVal,long lValE,int iReadE,int iWriteE,int iExceptE)
{
	long lVal;
	int iRead,iWrite,iExcept;

	FD_SET(s,pfdsRead);
	FD_SET(s,pfdsWrite);
	FD_SET(s,pfdsExcept);
	lVal = select(0,pfdsRead,pfdsWrite,pfdsExcept,psTimeVal);
	iRead = FD_ISSET(s,pfdsRead);
	iWrite = FD_ISSET(s,pfdsWrite);
	iExcept = FD_ISSET(s,pfdsExcept);
	if (lVal != lValE || iRead != iReadE || iWrite != iWriteE || iExcept != iExceptE)
		Fail("&returned: %ld%s%s%s expected: %ld%s%s%s",lVal,
			iRead ? " read" : "",
			iWrite ? " write" : "",
			iExcept ? " except" : "",
			lValE,
			iReadE ? " read" : "",
			iWriteE ? " write" : "",
			iExceptE ? " except" : "");
}

struct option_strings {
	int iOption;
	char *pszOption;
} sOptionStrings[] = {
	{SO_DEBUG,"SO_DEBUG"},
	{SO_ACCEPTCONN,"SO_ACCEPTCONN"},
	{SO_REUSEADDR,"SO_REUSEADDR"},
	{SO_KEEPALIVE,"SO_KEEPALIVE"},
	{SO_DONTROUTE,"SO_DONTROUTE"},
	{SO_BROADCAST,"SO_BROADCAST"},
	{SO_USELOOPBACK,"SO_USELOOPBACK"},
	{SO_LINGER,"SO_LINGER"},
	{SO_OOBINLINE,"SO_OOBINLINE"},
	{SO_DONTLINGER,"SO_DONTLINGER"},
	{SO_SNDBUF,"SO_SNDBUF"},
	{SO_RCVBUF,"SO_RCVBUF"},
	{SO_SNDLOWAT,"SO_SNDLOWAT"},
	{SO_RCVLOWAT,"SO_RCVLOWAT"},
	{SO_SNDTIMEO,"SO_SNDTIMEO"},
	{SO_RCVTIMEO,"SO_RCVTIMEO"},
	{SO_ERROR,"SO_ERROR"},
	{SO_TYPE,"SO_TYPE"},
	{TCP_NODELAY + 0xf000,"TCP_NODELAY"},
	{0,0}
};

void TestSockopt(char *pszMode,int iLevel,int iOption,int iNoProto,int iOptValE)
{
	int i,iOptVal,iOptValTest,iOptValRestore,iOptLen = sizeof(int);
	int iErrorExpected = iNoProto == SOCK_STREAM ? SOCK_DGRAM :
								iNoProto == SOCK_DGRAM ? SOCK_STREAM : 0;
	char *pszOption = "unknown";
	struct option_strings *psOptionStrings;

	i = iOption + (iLevel == IPPROTO_TCP ? 0xf000 : 0);
	for (psOptionStrings = sOptionStrings;
				(iOptVal = psOptionStrings->iOption) != 0;++psOptionStrings)
	{
		if (i == iOptVal)
		{
			pszOption = psOptionStrings->pszOption;
			break;
		}
	}

	Start("%s getsockopt(%s)",pszMode,pszOption);
	if (getsockopt(soClient,iLevel,iOption,(char *)&iOptVal,&iOptLen) == SOCKET_ERROR)
	{
		if (!iNoProto || WSAGetLastError() != WSAENOPROTOOPT)
			Fail(0);
	}
	else if (iErrorExpected)
		Fail("&error expected");
	else
	{
		iOptValTest = iOptValE == -1 ? iOptVal : iOptValE;
		if (iOptVal != iOptValTest || iOptLen != sizeof(int))
			Fail("&OptVal,OptLen: %d,%d expected; %d,%d",
										iOptVal,iOptLen,iOptValE,sizeof(int));
	}
	End();
	if (iOption == SO_ACCEPTCONN || iOption == SO_ERROR || iOption == SO_TYPE)
		return;	// can't set
	if (iOptValE == -1)
		iOptValTest /= 2;
	else
		iOptValTest = !iOptValE;
	iOptValRestore = iOptVal;

	Start("%s setsockopt(%s)",pszMode,pszOption);
	if (setsockopt(soClient,iLevel,iOption,(char *)&iOptValTest,iOptLen) == SOCKET_ERROR)
	{
		if (!iNoProto || WSAGetLastError() != WSAENOPROTOOPT)
			Fail(0);
	}
	else if (iErrorExpected)
		Fail("&error expected");
	End();

	Start("%s getsockopt(%s)mod",pszMode,pszOption);
	if (getsockopt(soClient,iLevel,iOption,(char *)&iOptVal,&iOptLen) == SOCKET_ERROR)
	{
		if (!iNoProto || WSAGetLastError() != WSAENOPROTOOPT)
			Fail(0);
	}
	else if (iErrorExpected)
		Fail("&error expected");
	else
	{
		if (iOptVal != iOptValTest || iOptLen != sizeof(int))
			Fail("&OptVal,OptLen: %d,%d expected; %d,%d",
										iOptVal,iOptLen,iOptValTest,sizeof(int));
	}
	End();

	Start("%s setsockopt(%s)res",pszMode,pszOption);
	if (setsockopt(soClient,iLevel,iOption,(char *)&iOptValRestore,iOptLen) == SOCKET_ERROR)
	{
		if (!iNoProto || WSAGetLastError() != WSAENOPROTOOPT)
			Fail(0);
	}
	else if (iErrorExpected)
		Fail("&error expected");
	End();

	Start("%s getsockopt(%s)res",pszMode,pszOption);
	if (getsockopt(soClient,iLevel,iOption,(char *)&iOptVal,&iOptLen) == SOCKET_ERROR)
	{
		if (!iNoProto || WSAGetLastError() != WSAENOPROTOOPT)
			Fail(0);
	}
	else if (iErrorExpected)
		Fail("&error expected");
	else
	{
		if (iOptVal != iOptValRestore || iOptLen != sizeof(int))
			Fail("&OptVal,OptLen: %d,%d expected; %d,%d",
										iOptVal,iOptLen,iOptValRestore,sizeof(int));
	}
	End();
}

DWORD dwClientHandlerAddr;
DWORD dwServerHandlerAddr;

int TestSocket(int iMode)
{
	SOCKADDR_IN sPeerAddr;
	int i, iRead, iWritten, iLen, iOptLen;
	char *pszMode = iMode == SOCK_STREAM ? "STREAM" : "DATA_GRAM";
	BYTE bReadFailed = 0;
	char sz[80];
	u_long ulArg;
	fd_set fdsRead, fdsWrite, fdsExcept;
	struct timeval sTimeVal;
	struct linger sLinger;
	long lVal;

	if (dwServerIPAddress == dwMyIPAddress)
		StartServer(iMode);

	Start("socket(INVALID_AF,%s)",pszMode);
	if ((soClient = socket(1000,iMode,0)) != INVALID_SOCKET ||
					WSAGetLastError() != WSAEAFNOSUPPORT)
		Fail(0);
	End();
	Start("socket(INVALID_TYPE)");
	if ((soClient = socket(AF_INET,1000,0)) != INVALID_SOCKET ||
					WSAGetLastError() != WSAESOCKTNOSUPPORT)
		Fail(0);
	End();
	Start("socket(%s,INVALID_PROTOCOL)",pszMode);
	if ((soClient = socket(AF_INET,iMode,1000)) != INVALID_SOCKET ||
					WSAGetLastError() != WSAEPROTONOSUPPORT)
		Fail(0);
	End();
	Start("socket(%s,INVALID_PROTOCOL)",pszMode);
	if ((soClient = socket(AF_INET,iMode,1000)) != INVALID_SOCKET ||
					WSAGetLastError() != WSAEPROTONOSUPPORT)
		Fail(0);
	End();
	// Conflicting type and protocol
	Start("socket(AF_INET,%s,%s)",pszMode,
						iMode == SOCK_STREAM ? "IPPROTO_UDP" : "IPPROTO_TCP");
	if ((soClient = socket(AF_INET,iMode,
		iMode == SOCK_STREAM ? IPPROTO_UDP : IPPROTO_TCP)) != INVALID_SOCKET ||
#ifdef _Windows
					WSAGetLastError() != WSAEPROTONOSUPPORT)	// I think this is a bug in Winsock
#else
					WSAGetLastError() != WSAEPROTOTYPE)
#endif
		Fail(0);
	End();
	// now for a valid call
	Start("socket(%s)", pszMode);
	if ((soClient = socket(AF_INET,iMode,0)) == INVALID_SOCKET)
		Fail(0);
	End();

	TestSockopt(pszMode,SOL_SOCKET,SO_ACCEPTCONN,iMode & SOCK_DGRAM,0);
	TestSockopt(pszMode,SOL_SOCKET,SO_BROADCAST,iMode & SOCK_STREAM,0);
	TestSockopt(pszMode,SOL_SOCKET,SO_DEBUG,0,0);
	TestSockopt(pszMode,SOL_SOCKET,SO_DONTLINGER,iMode & SOCK_DGRAM,1);
	TestSockopt(pszMode,SOL_SOCKET,SO_ACCEPTCONN,iMode & SOCK_DGRAM,0);
	TestSockopt(pszMode,SOL_SOCKET,SO_ERROR,0,0);
	TestSockopt(pszMode,SOL_SOCKET,SO_KEEPALIVE,iMode & SOCK_DGRAM,0);
	TestSockopt(pszMode,SOL_SOCKET,SO_OOBINLINE,iMode & SOCK_DGRAM,0);
	TestSockopt(pszMode,SOL_SOCKET,SO_REUSEADDR,0,0);
	TestSockopt(pszMode,SOL_SOCKET,SO_TYPE,0,iMode);
	TestSockopt(pszMode,IPPROTO_TCP,TCP_NODELAY,iMode & SOCK_DGRAM,0);
	TestSockopt(pszMode,SOL_SOCKET,SO_SNDBUF,0,-1);
	TestSockopt(pszMode,SOL_SOCKET,SO_SNDBUF,0,-1);

	Start("%s getsockopt(SO_LINGER)",pszMode);
	iOptLen = sizeof(struct linger);
	if (getsockopt(soClient,SOL_SOCKET,SO_LINGER,(char *)&sLinger,&iOptLen) == SOCKET_ERROR)
	{
		if (iMode != SOCK_DGRAM || WSAGetLastError() != WSAENOPROTOOPT)
			Fail(0);
	}
	else if (iMode == SOCK_DGRAM)
		Fail("&error expected");
	else if (sLinger.l_onoff || sLinger.l_linger || iOptLen != sizeof(struct linger))
			Fail("&l_onoff,l_linger,len: %d,%d,%d expected; 0,0,%d",
										sLinger.l_onoff,sLinger.l_linger,sizeof(struct linger));
	End();

	// make soClient non-blocking
	Start("%s ioctlsocket(FIONBIO)", pszMode);
	ulArg = 1;
	if (ioctlsocket(soClient,FIONBIO,&ulArg) == SOCKET_ERROR)
		Fail(0);
	End();

	Start("%s bind()",pszMode);
	sClientAddr.sin_family = AF_INET;
	sClientAddr.sin_port = 0;
	sClientAddr.sin_addr.s_addr = INADDR_ANY;
	if (bind(soClient,(PSOCKADDR)&sClientAddr,sizeof(SOCKADDR)) == SOCKET_ERROR)
		Fail(0);
	End();

	Start("%s connect()",pszMode);
	sClientAddr.sin_port = htons(wPort);
	sClientAddr.sin_addr.s_addr = dwServerIPAddress;
	if (connect(soClient,(PSOCKADDR)&sClientAddr,sizeof(SOCKADDR)) == SOCKET_ERROR)
	{
		if (WSAGetLastError() != WSAEWOULDBLOCK)
			Fail(0);
	}
	End();

	Start("%s FD_ISSET() after FD_ZERO()",pszMode);
	FD_ZERO(&fdsRead);
	FD_ZERO(&fdsWrite);
	FD_ZERO(&fdsExcept);
	if (FD_ISSET(soClient,&fdsRead))
		Fail(0);
	End();

	Start("%s FD_ISSET() after FD_SET()",pszMode);
	FD_SET(soClient,&fdsRead);
	if (!FD_ISSET(soClient,&fdsRead))
		Fail(0);
	End();

	Start("%s FD_ISSET() after FD_CLR()",pszMode);
	FD_CLR(soClient,&fdsRead);
	if (FD_ISSET(soClient,&fdsRead))
		Fail(0);
	End();

	SetSecondsTimer(wTimeOut);
	Start("%s recv()",pszMode);
	while ((iRead = recv(soClient,arcReadBuf,sizeof(arcReadBuf),0)) < 0)
	{
		if (WSAGetLastError() == WSAEWOULDBLOCK)
			break;
		if (WSAGetLastError() != WSAENOTCONN)
		{
			Fail(0);
			break;
		}
		if (DiagWait())
		{
			iFail = 1;
			break;
		}
	}
	if (iRead >= 0)
		Fail("&unexpected count %d",iRead);
	End();

	Start("%s select(writefds)==1",pszMode);
	FD_SET(soClient,&fdsWrite);
	sTimeVal.tv_sec = 0;
	sTimeVal.tv_usec = 0;
	if ((lVal = select(0,0,&fdsWrite,0,&sTimeVal)) != 1)
		Fail("&returned: %ld expected 1",lVal);
	else if (!FD_ISSET(soClient,&fdsWrite))
		Fail(0);
	End();

	Start("%s select(readfds)==0",pszMode);
	FD_SET(soClient,&fdsRead);
	if ((lVal = select(0,&fdsRead,0,0,&sTimeVal)) != 0)
		Fail("&returned: %ld expected 0",lVal);
	else if (FD_ISSET(soClient,&fdsRead))
		Fail(0);
	End();

	Start("%s select(exceptfds)==0",pszMode);
	FD_SET(soClient,&fdsExcept);
	if ((lVal = select(0,0,0,&fdsExcept,&sTimeVal)) != 0)
		Fail("&returned: %ld expected 0",lVal);
	else if (FD_ISSET(soClient,&fdsExcept))
		Fail(0);
	End();

	Start("%s getpeername()",pszMode);
	iLen = sizeof(SOCKADDR);
	if (getpeername(soClient,(PSOCKADDR)&sPeerAddr,&iLen) == SOCKET_ERROR)
		Fail(0);
	if (sPeerAddr.sin_addr.s_addr != dwServerIPAddress)
		Fail("&returned wrong RemoteHost %s",IPAddressString(sPeerAddr.sin_addr.s_addr));
	if (sPeerAddr.sin_port != htons(wPort))
		Fail("&returned wrong RemotePort %d",ntohs(sPeerAddr.sin_port));
	End();


	// Send some data to echo server
	Start("%s send",pszMode);
	if ((iWritten = send(soClient, szData, sizeof(szData), 0)) == SOCKET_ERROR)
		Fail(0);
	else if (iWritten != sizeof(szData))
		Fail("&short write: %d expected %d",iWritten,sizeof(szData));
	End();

	Start("%s select(readfds)==wait",pszMode);
	SetSecondsTimer(wTimeOut);
	while (1)
	{
		FD_SET(soClient,&fdsRead);
		if ((lVal = select(0,&fdsRead,0,0,&sTimeVal)) == 1)
		{
			if (FD_ISSET(soClient,&fdsRead))
				break;
			Fail(0);
		}
		else if (lVal != 0)
		{
			Fail("&returned: %ld expected 0 or 1");
			break;
		}
		if (DiagWait())
		{
			iFail = 1;
			break;
		}
	}
	End();

	// check for queued data
	Start("%s ioctlsocket(FIONREAD)", pszMode);
	ulArg = -1L;
	if (ioctlsocket(soClient,FIONREAD,&ulArg) == SOCKET_ERROR)
		Fail(0);
	if (ulArg != sizeof(szData))
		Fail("&received %lu expected %u",ulArg,sizeof(szData));
	End();

	// Send some more data to echo server
	Start("%s send()",pszMode);
	if ((iWritten = send(soClient, szData, sizeof(szData), 0)) < 0)
		Fail(0);
	else if (iWritten != sizeof(szData))
		Fail("&short write: %d expected %d",iWritten,sizeof(szData));
	End();

	Delay(110);		// give time to receive response

	// check for queued data
	// Note that the documentation and implementation differ in the case
	// of SOCK_DATAGRAM between Winsock 2 and the Winsock 1.1 spec. The test
	// works for both cases.
	Start("%s ioctlsocket(FIONREAD)", pszMode);
	ulArg = -1L;
	if (ioctlsocket(soClient,FIONREAD,&ulArg) == SOCKET_ERROR)
		Fail(0);
	if (ulArg != 2 * sizeof(szData) && (iMode == SOCK_STREAM || (ulArg != sizeof(szData))))
		Fail("&received %lu expected %u",ulArg,2 * sizeof(szData));
	End();

	// Read back without de-queueing data
	Start("%s recv(pBuf,MSG_PEEK)==%d", pszMode, sizeof(szData));
	memset(arcReadBuf,0,sizeof(arcReadBuf));
	if ((iRead = ReadSocketCount(soClient, iMode, arcReadBuf,sizeof(szData),sizeof(szData),MSG_PEEK)) == sizeof(szData))
	{
		if (memcmp(szData,arcReadBuf,iRead) != 0)
			Fail("Readback error:\n%.*s\n%.*s",iRead,szData,iRead,arcReadBuf);
	}
	End();

	// Read data
	Start("%s recv(pBuf,0)==%d", pszMode, sizeof(szData));
	memset(arcReadBuf,0,sizeof(arcReadBuf));
	if ((iRead = ReadSocketCount(soClient, iMode, arcReadBuf,sizeof(szData),
												sizeof(szData),0)) == sizeof(szData))
	{
		if (memcmp(szData,arcReadBuf,iRead) != 0)
			Fail("Readback error:\n%.*s\n%.*s",iRead,szData,iRead,arcReadBuf);
	}
	End();

	// Read data
	Start("%s recv(pBuf,0)==%d", pszMode, sizeof(szData));
	memset(arcReadBuf,0,sizeof(arcReadBuf));
	if ((iRead = ReadSocketCount(soClient,iMode,arcReadBuf,sizeof(arcReadBuf),
												sizeof(szData),0)) == sizeof(szData))
	{
		if (memcmp(szData,arcReadBuf,iRead) != 0)
			Fail("Readback error:\n%.*s\n%.*s",iRead,szData,iRead,arcReadBuf);
	}
	if (iRead == -1)
		bReadFailed = 1;
	End();

	// next read should block
	Start("%s recv(pBuf,0)", pszMode);
	if ((iRead = recv(soClient, arcReadBuf,sizeof(arcReadBuf),0)) != -1)
		Fail("&unexpected data count %d", iRead);
	else if (WSAGetLastError() != WSAEWOULDBLOCK)
		Fail(0);
	End();

	Start("%s select(readfds,writefds,exceptfds)==1",pszMode);
	TestSelect(soClient,&fdsRead,&fdsWrite,&fdsExcept,&sTimeVal,1,0,1,0);
	End();

	if (iMode == SOCK_STREAM)
	{
		Start("shutdown(1)");
		if (shutdown(soClient,1) == SOCKET_ERROR)
			Fail(0);
		End();

		Start("%s select(readfds,writefds,exceptfds)==1,2",pszMode);
		SetSecondsTimer(wTimeOut);
		sTimeVal.tv_usec = 500000L;
		while (1)
		{
			FD_SET(soClient,&fdsRead);
			FD_SET(soClient,&fdsWrite);
			FD_SET(soClient,&fdsExcept);
			lVal = select(0,&fdsRead,&fdsWrite,&fdsExcept,&sTimeVal);
			if (lVal == 1)
			{
				if (FD_ISSET(soClient,&fdsRead))
				{
					Fail("&unexpected readfds");
					break;
				}
				else if (FD_ISSET(soClient,&fdsExcept))
				{
					Fail("&unexpected exceptfds");
					break;
				}
				else if (!FD_ISSET(soClient,&fdsWrite))
				{
					Fail("&expected writefds");
					break;
				}
			}
			else if (lVal == 2)
			{
				if (FD_ISSET(soClient,&fdsExcept))
					Fail("&unexpected exceptfds");
				else if (!FD_ISSET(soClient,&fdsRead))
					Fail("&expected readfds");
				else if (!FD_ISSET(soClient,&fdsWrite))
					Fail("&expected writefds");
				break;
			}
			else
			{
				Fail("&returned: %ld expected 1 or 2",lVal);
				break;
			}
			if (DiagWait())
			{
				iFail = 1;
				break;
			}
		}
		End();

		TestSockopt(pszMode,SOL_SOCKET,SO_ERROR,0,0);

		Start("%s recv() with peer closing connection",pszMode);
		if ((iRead = recv(soClient,arcReadBuf,sizeof(arcReadBuf),0)) < 0)
			Fail(0);
		else if (iRead > 0)
			Fail("&unexpected count %d",iRead);
		End();

		sTimeVal.tv_usec = 0;
		Start("%s select(readfds,writefds,exceptfds)==2",pszMode);
		TestSelect(soClient,&fdsRead,&fdsWrite,&fdsExcept,&sTimeVal,2,1,1,0);
		End();
	}

	Start("%s closesocket(client)",pszMode);
	if (closesocket(soClient) == SOCKET_ERROR)
		Fail(0);
	End();

	// Test blocking socket with select()
	Start("socket(%s)b", pszMode);
	if ((soClient = socket(AF_INET,iMode,0)) == INVALID_SOCKET)
		Fail(0);
	End();

	Start("%s bind()b",pszMode);
	sClientAddr.sin_family = AF_INET;
	sClientAddr.sin_port = 0;
	sClientAddr.sin_addr.s_addr = INADDR_ANY;
	if (bind(soClient,(PSOCKADDR)&sClientAddr,sizeof(SOCKADDR)) == SOCKET_ERROR)
		Fail(0);
	End();

	Start("%s connect()b",pszMode);
	sClientAddr.sin_port = htons(wPort);
	sClientAddr.sin_addr.s_addr = dwServerIPAddress;
	if (connect(soClient,(PSOCKADDR)&sClientAddr,sizeof(SOCKADDR)) == SOCKET_ERROR)
		Fail(0);
	End();

	Start("%s select(readfds,writefds,exceptfds)b==1",pszMode);
	TestSelect(soClient,&fdsRead,&fdsWrite,&fdsExcept,&sTimeVal,1,0,1,0);
	End();

	// Send some data to echo server
	Start("%s send()b",pszMode);
	if ((iWritten = send(soClient, szData, sizeof(szData), 0)) == SOCKET_ERROR)
		Fail(0);
	else if (iWritten != sizeof(szData))
		Fail("&short write: %d expected %d",iWritten,sizeof(szData));
	End();

	Start("%s select(readfds)b==wait",pszMode);
	SetSecondsTimer(wTimeOut);
	while (1)
	{
		FD_SET(soClient,&fdsRead);
		if ((lVal = select(0,&fdsRead,0,0,&sTimeVal)) == 1)
		{
			if (FD_ISSET(soClient,&fdsRead))
				break;
			Fail(0);
		}
		else if (lVal != 0)
		{
			Fail("&returned: %ld expected 0 or 1");
			break;
		}
		if (DiagWait())
		{
			iFail = 1;
			break;
		}
	}
	End();

	// check for queued data
	Start("%s ioctlsocket(FIONREAD)b", pszMode);
	ulArg = -1L;
	if (ioctlsocket(soClient,FIONREAD,&ulArg) == SOCKET_ERROR)
		Fail(0);
	if (ulArg != sizeof(szData))
		Fail("&received %lu expected %u",ulArg,sizeof(szData));
	End();

	if (ulArg)
	{				// only read if something ias available
		// Read back without de-queueing data
		Start("%s recv(pBuf,MSG_PEEK)b==%d", pszMode, sizeof(szData));
		memset(arcReadBuf,0,sizeof(arcReadBuf));
		if ((iRead = ReadSocketCount(soClient, iMode, arcReadBuf,sizeof(szData),sizeof(szData),MSG_PEEK)) == sizeof(szData))
		{
			if (memcmp(szData,arcReadBuf,iRead) != 0)
				Fail("Readback error:\n%.*s\n%.*s",iRead,szData,iRead,arcReadBuf);
		}
		End();

		// Read data
		Start("%s recv(pBuf,0)b==%d", pszMode, sizeof(szData));
		memset(arcReadBuf,0,sizeof(arcReadBuf));
		if ((iRead = ReadSocketCount(soClient, iMode, arcReadBuf,sizeof(szData),
													sizeof(szData),0)) == sizeof(szData))
		{
			if (memcmp(szData,arcReadBuf,iRead) != 0)
				Fail("Readback error:\n%.*s\n%.*s",iRead,szData,iRead,arcReadBuf);
		}
		End();
	}
	Start("%s select(readfds,writefds,exceptfds)b==1",pszMode);
	TestSelect(soClient,&fdsRead,&fdsWrite,&fdsExcept,&sTimeVal,1,0,1,0);
	End();

	if (iMode == SOCK_STREAM && dwServerIPAddress == dwMyIPAddress && bServerInUse)
	{
		// test select after reset
		for (bReset = 1;bReset;)
			PollServer();
		Start("%s select()reset==wait",pszMode);
		SetSecondsTimer(wTimeOut);
		while (1)
		{
			FD_SET(soClient,&fdsRead);
			FD_SET(soClient,&fdsWrite);
			FD_SET(soClient,&fdsExcept);
			lVal = select(0,&fdsRead,&fdsWrite,&fdsExcept,&sTimeVal);
			if (lVal && FD_ISSET(soClient,&fdsRead))
			{
				int i;

				End();
				// try to figure out what SO_ERROR actually does
				TestSockopt(pszMode,SOL_SOCKET,SO_ERROR,0,0);
				Start("recv after reset");
				if ((i = recv(soClient,arcReadBuf,sizeof(arcReadBuf),0)) < 0)
				{
					if (WSAGetLastError() != WSAECONNRESET)
						Fail(0);
					End();
					TestSockopt(pszMode,SOL_SOCKET,SO_ERROR,0,0);
					goto done;
				}
				Fail("&unexpected receive of %ld",i);
				End();
				Start("%s select()reset==wait",pszMode);
			}
			if (DiagWait())
			{
				iFail = 1;
				break;
			}
		}
		End();
	}
	done:
	Start("%s closesocket(client)b",pszMode);
	if (closesocket(soClient) == SOCKET_ERROR)
		Fail(0);
	End();

	// Volume data transfers
	if (!bReadFailed)
	{
		for (iBlocking = 0;iBlocking < 2;++iBlocking)
		{
			for (i = 0;i < BUFLEN;++i)
				rgcWriteBuf[i] = (char)i;
			Start("socket(%s)v", pszMode);
			if ((soClient = socket(PF_INET,iMode,0)) == INVALID_SOCKET)
				Fail(0);
			End();
			SetupBulk(iMode,pszMode);
			if (!iBlocking)
			{
				// make soClient non-blocking
				Start("%s ioctlsocket(FIONBIO)v", pszMode);
				ulArg = 1;
				if (ioctlsocket(soClient,FIONBIO,&ulArg) == SOCKET_ERROR)
					Fail(0);
				End();
			}

			Start("%s Write/Read bulk %s",pszMode,iBlocking ? "Blocking" : "");
			while (bWriting || dwBufsRead != dwBufsWritten)
			{
				if (dwBufsWritten - dwBufsRead < WRITE_AHEAD && bWriting)
				{
					if (WriteBlock(soClient) < 0)
					{
						Fail("&write error");
						break;
					}
				}
				if ((i = ReadBlock(soClient)) < 0)
				{
					if (i == -2)
						Fail("&read error");
					else if (!iErrorAddress)
						Fail(" Readback error,  written %lu read %lu offset %u good %lu bad %lu",
								dwBufsWritten,dwBufsRead + 1,iErrorAddress,
								dwBufsRead,*(DWORD *)arcReadBuf);
					else
						Fail(" Readback error,  written %lu read %lu offset %u good x%x bad x%x",
								dwBufsWritten,dwBufsRead + 1,iErrorAddress,
								(BYTE)iErrorAddress,(BYTE)arcReadBuf[i]);
					break;
				}
				else if (i == 0)	// nothing read
				{
					if ((lTimeRead - GetMilliSeconds()) <= 0)
					{
						Fail(" Read timeout, written %lu read %lu blocks",
													dwBufsWritten,dwBufsRead);
						break;
					}
				}
			}
			if ((lTime = GetMilliSeconds() - lTime) == 0)
				lTime = 1;
			sprintf(sz,"%.0f",(FACTOR * BUFLEN * dwBufsWritten) / lTime);
			Seperate1000(sz);
			printf("%-9s%s: %14s bytes/sec\n",pszMode,
								iBlocking ? " Blocking" : "         ",sz);
			End();
			Start("%s closesocket(client)v",pszMode);
			if (closesocket(soClient) < 0)
				Fail(0);
			End();
		}
	}
	else
		printf("Volume traffic tests skipped\n");

	if (dwServerIPAddress == dwMyIPAddress)
		StopServer();
	return 0;
}

#pragma argsused
void SetupBulk(int iMode, char *pszMode)
{
	int i = 1000;
	// Blocking connect
	sClientAddr.sin_port = htons(wPort);
	//sClientAddr.wLocalPort = 0;
	Start("connect(%s)",pszMode);
	if (connect(soClient,(PSOCKADDR)&sClientAddr,sizeof(SOCKADDR)) == SOCKET_ERROR)
		Fail(0);
	End();
	while (i--)
		PollServer();
	// Write and read back data for 10 seconds
	lTime = GetMilliSeconds();
	while (lTime == GetMilliSeconds());
	lTime = GetMilliSeconds();
	lTarget = (lTime + iBulkTime);					// 10 seconds default
	lTimeRead = (GetMilliSeconds() + iBulkTime / 2);	// 5 seconds default
	dwBufsWritten = dwBufsRead = 0;
	iWriteSize = iReadSize = BUFLEN;
	iWriteOffset = iReadOffset = iBlockError = 0;
	bWriting = 1;
	wRead = wWrite = 0;
}

// Write block for bulk transfer
#pragma argsused
int WriteBlock(SOCKET soSock)
{
	int iWritten;
	fd_set fdsWrite;
	struct timeval sTimeVal;

	if (iBlocking)
	{
		FD_ZERO(&fdsWrite);
		FD_SET(soSock,&fdsWrite);
		sTimeVal.tv_sec = 0;
		sTimeVal.tv_usec = 0;
		if (select(0,0,&fdsWrite,0,&sTimeVal) != 1)
			return 0;
	}
	*(DWORD *)rgcWriteBuf = dwBufsWritten;
	if ((iWritten = send(soSock,rgcWriteBuf + iWriteOffset,
									iWriteSize, 0)) < 0)
	{
		if (WSAGetLastError() != WSAEWOULDBLOCK)
		{
			return -1;
		}
	}
	else if (iWritten < iWriteSize)
	{
		iWriteOffset += iWritten;
		iWriteSize -= iWritten;
	}
	else
	{
		iWriteOffset = 0;
		iWriteSize = BUFLEN;
		dwBufsWritten++;
		if ((int)(lTarget - GetMilliSeconds()) <= 0)
			bWriting = 0;
	}
	return 0;
}

// Read block for bulk transfer
#pragma argsused
int ReadBlock(SOCKET soSock)
{
	int iRead, i;
	fd_set fdsRead;
	struct timeval sTimeVal;

	PollServer();

	if (iBlocking)
	{
		FD_ZERO(&fdsRead);
		FD_SET(soSock,&fdsRead);
		sTimeVal.tv_sec = 0;
		sTimeVal.tv_usec = 0;
		if (select(0,&fdsRead,0,0,&sTimeVal) != 1)
			return 0;
	}
	if ((iRead = recv(soSock,arcReadBuf + iReadOffset,
									iReadSize, 0)) < 0)
	{
		if (WSAGetLastError() != WSAEWOULDBLOCK)
		{
			return -2;
		}
		return 0;	// nothing read
	}
	else if (iRead < iReadSize)
	{
		iReadOffset += iRead;
		iReadSize -= iRead;
		lTimeRead = (GetMilliSeconds() + 5000);	// 5 seconds
	}
	else
	{
		iReadOffset = 0;
		iReadSize = BUFLEN;
		lTimeRead = (GetMilliSeconds() + 5000);	// 5 seconds
		if (*(DWORD *)arcReadBuf != dwBufsRead)
		{
			iErrorAddress = 0;
			return -3;
		}
		for (i = 4;i < BUFLEN;++i)
		{
			if (arcReadBuf[i] != (char)i)
			{
				iErrorAddress = i;
				return -3;
			}
		}
		dwBufsRead++;
	}
	return 1;
}

// Seperate 1000's
void Seperate1000(char *psz)
{
	int i, iChars = strlen(psz);
	int iSpaces = (iChars - 1) / 3;

	psz[iChars + iSpaces] = 0;
	for (i = 0;iChars--;)
	{
		psz[iChars + iSpaces] = psz[iChars];
		if (++i == 3)
		{
			i = 0;
			--iSpaces;
			if (iChars + iSpaces > 0)
				psz[iChars + iSpaces] = ' ';
		}
	}
}

// Read expected number of bytes or until interrupted
int ReadSocketCount(int iSocket,int iMode,char *pBuf,int iBufLen,int iExpectedLen,
	WORD wFlags)
{
	int i, iRead = 0;
	char *pReadBuf = pBuf;
	SetSecondsTimer(wTimeOut);
	while (1)
	{
		if (DiagWait())
		{
			iFail = 1;
			if (iVerbose & 2)
				printf("Read %d\n",iRead);
			return -1;
		}
		if ((i = recv(iSocket,pReadBuf,iBufLen - iRead,wFlags)) < 0)
		{
			if (WSAGetLastError() != WSAEWOULDBLOCK)
			{
				Fail(0);
				return -1;
			}
		}
		else if (i == 0 && iMode == SOCK_STREAM) {
			Fail("Peer closed connection while reading %d byte(s)",
							iExpectedLen);
			return -1;
		}
		else {
			if (wFlags & MSG_PEEK)
				iRead = i;
			else
				iRead += i;
			if (iRead >= iExpectedLen)
				break;
			if (pReadBuf && !(wFlags & MSG_PEEK))
				pReadBuf += i;
		}
	}
	if (iRead != iExpectedLen)
		Fail("&unexpected count %d, expected %d",iRead,iExpectedLen);
	return iRead;
}

// Local echo server

int iServerMode;

#define NUM_QUEUED	3		// number of Listen/Accept queued entries
int StartServer(int iMode)
{
	int retv = 0,i;
	SOCKET arsoClient[NUM_QUEUED + 1],arsoServer[NUM_QUEUED + 1];
	char *pszMode = iMode == SOCK_STREAM ? "STREAM" : "DATA_GRAM";
	u_long ulArg;
	SOCKADDR_IN sSockAddr;
	fd_set fdsRead, fdsWrite, fdsExcept;
	struct timeval sTimeVal = {0,0};

	FD_ZERO(&fdsRead);
	FD_ZERO(&fdsWrite);
	FD_ZERO(&fdsExcept);
	iServerMode = iMode;
	Start("%s server socket()",pszMode);
	if ((soListen = socket(PF_INET,iMode,0)) == INVALID_SOCKET)
	{
		Fail(0);
		retv = -1;
	}
	End();
	// make soServer non-blocking
	Start("%s server ioctlsocket(FIONBIO)", pszMode);
	ulArg = 1;
	if (ioctlsocket(soListen,FIONBIO,&ulArg) == SOCKET_ERROR)
	{
		Fail(0);
		retv = -2;
	}
	End();
	memset(&sSockAddr,0,sizeof(sSockAddr));
	sSockAddr.sin_family = AF_INET;
	sSockAddr.sin_port = htons(wPort);
	sSockAddr.sin_addr.s_addr = dwServerIPAddress;

	Start("%s server bind()",pszMode);
	if (bind(soListen,(SOCKADDR *)&sSockAddr,sizeof(SOCKADDR)) == SOCKET_ERROR)
		Fail(0);
	End();
	if (iMode == SOCK_DGRAM)
	{
		soServer = soListen;
		soListen = INVALID_SOCKET;
		Start("%s server connect()",pszMode);
		sSockAddr.sin_port = 0;
		if (connect(soServer,(PSOCKADDR)&sSockAddr,sizeof(SOCKADDR)) == SOCKET_ERROR)
		{
			Fail(0);
			End();
			return -3;
		}
		End();
		Start("%s server select(0)",pszMode);
		TestSelect(soServer,&fdsRead,&fdsWrite,&fdsExcept,&sTimeVal,1,0,1,0);
		End();
	}
	else
	{
		Start("%s server listen()",pszMode);
		sServerAddr.sin_port = 0;
		sServerAddr.sin_addr.s_addr = 0;
		if (listen(soListen, NUM_QUEUED) == SOCKET_ERROR)
		{
			Fail(0);
			End();
			return -3;
		}
		End();

		Start("%s server select(0)",pszMode);
		TestSelect(soListen,&fdsRead,&fdsWrite,&fdsExcept,&sTimeVal,0,0,0,0);
		End();

		for (i = 0;i <= NUM_QUEUED;++i)
		{
			Start("%s client %d socket()",pszMode,i);
			if ((arsoClient[i] = socket(PF_INET,iMode,0)) == INVALID_SOCKET)
				Fail(0);
			End();
			Start("%s client %d connect()",pszMode,i);
			sSockAddr.sin_addr.s_addr = dwServerIPAddress;
			if (connect(arsoClient[i],(PSOCKADDR)&sSockAddr,sizeof(SOCKADDR)) == SOCKET_ERROR)
			{
				if (i < NUM_QUEUED || WSAGetLastError() != WSAECONNREFUSED)
					Fail(0);
			}
			else if (i == NUM_QUEUED)
				Fail(0);		// should have failed!
			End();
			if (i == NUM_QUEUED)
			{
				Start("%s client %d closesocket()",pszMode,i);
				if (closesocket(arsoClient[NUM_QUEUED]) == SOCKET_ERROR)
					Fail(0);
				End();
			}
		}
		Start("%s server select(1)",pszMode);
		TestSelect(soListen,&fdsRead,&fdsWrite,&fdsExcept,&sTimeVal,1,1,0,0);
		End();
		for (i = 0;i <= NUM_QUEUED;++i)
		{
			Start("%s server %d accept()",pszMode,i);
			if ((arsoServer[i] = accept(soListen,(PSOCKADDR)&sServerAddr,0)) == INVALID_SOCKET)
			{
				if (i < NUM_QUEUED || WSAGetLastError() != WSAEWOULDBLOCK)
					Fail(0);
			}
			else if (i == NUM_QUEUED)
			{
				Fail(0);		// should have failed!
				closesocket(arsoServer[NUM_QUEUED]);
			}
			End();
		}
		Start("%s server select(2)",pszMode);
		TestSelect(soListen,&fdsRead,&fdsWrite,&fdsExcept,&sTimeVal,0,0,0,0);
		End();
		for (i = 0;i < NUM_QUEUED;++i)
		{
			Start("%s client %d closesocket(%u)",pszMode,i,arsoClient[i]);
			if (closesocket(arsoClient[i]) == SOCKET_ERROR)
				Fail(0);
			End();
			Start("%s server %d closesocket(%u)",pszMode,i,arsoServer[i]);
			if (closesocket(arsoServer[i]) == SOCKET_ERROR)
				Fail(0);
			End();
		}
	}
	bServerInUse = 1;
	return retv;
}

char arcBuf[2 * BUFLEN];

void PollServer(void)
{
	int iCount,iLen;
	SOCKADDR_IN sAddr;

	if (!bServerInUse)
		return;
	//only echo data back
	if (iServerMode == SOCK_STREAM)
	{
		if (soServer == INVALID_SOCKET)
		{
			if ((soServer = accept(soListen,0,0)) == INVALID_SOCKET)
			{
				if (WSAGetLastError() != WSAEWOULDBLOCK)
					ServerFail("accept()");
				return;
			}
			else if (bDebugServer)
				printf("Stream Server accepted connection\n");
		}

		if ((iCount = recv(soServer, arcBuf, sizeof(arcBuf), 0)) > 0)
		{
			if (bDebugServer)
				printf("Stream Server receive %d\n",iCount);
			if (send(soServer, arcBuf, iCount, 0) != iCount)
				ServerFail("send()");
		}
		else if (iCount == 0)
		{
			if (shutdown(soServer,2) != 0)
				ServerFail("shutdown()");
			if (bDebugServer)
				printf("Stream Server closed connection\n");
			if (closesocket(soServer) == SOCKET_ERROR)
				ServerFail("closesocket()");
			soServer = INVALID_SOCKET;
		}
		else if (WSAGetLastError() != WSAEWOULDBLOCK)
			ServerFail("recv()");
		if (bReset)
		{
			struct linger sLinger = {1,0};

			Start("Server setsockopt(SO_LINGER)");
			if (setsockopt(soServer,SOL_SOCKET,SO_LINGER,(char *)&sLinger,
												sizeof(struct linger)) == SOCKET_ERROR)
				Fail(0);
			End();
			if (bDebugServer)
				printf("Stream Server resetting connection\n");
			Start("Server closesocket()reset");
			if (closesocket(soServer) == SOCKET_ERROR)
				Fail(0);
			End();
			soServer = INVALID_SOCKET;
			bReset = 0;
		}
	}
	else
	{
		iLen = sizeof(SOCKADDR);
		if ((iCount = recvfrom(soServer,arcBuf,sizeof(arcBuf),0,
					(SOCKADDR *)&sAddr,&iLen)) > 0)
		{
			if (bDebugServer)
				printf("Datagram Server receive %d\n",iCount);
			if (sendto(soServer,arcBuf,iCount,0,(SOCKADDR *)&sAddr,iLen) != iCount)
				ServerFail("sendto()");
		}
		else if (WSAGetLastError() != WSAEWOULDBLOCK)
			ServerFail("recvfrom()");
	}
}

void StopServer(void)
{
	bServerInUse = 0;
	if (soListen != INVALID_SOCKET)
	{
		Start("Listen closesocket()");
		if (closesocket(soListen) == SOCKET_ERROR)
			Fail(0);
		End();
		soListen = INVALID_SOCKET;
	}
	if (soServer != INVALID_SOCKET)
	{
		Start("Server closesocket()");
		if (closesocket(soServer) == SOCKET_ERROR)
			Fail(0);
		End();
		soServer = INVALID_SOCKET;
	}
}

typedef struct sErr {
	int iErr;
	char *pszErr;
} SERROR;

extern SERROR sError[];

/*
Create a human understandable string from a Sockets error code.

Argument:
	uErrCode - The sockets error code.

Returns:
	A pointer to the (static) string representation of the error.
*/
char *GetErrorString(int iError)
{
	int i;

	for (i = 0;sError[i].iErr != 0x7fff;++i) {
		if (iError == sError[i].iErr)
			break;
	}
	if (sError[i].iErr == 0x7fff)
	{
		static char szErr[60];
		sprintf(szErr,"Unknown error 0x%x %u",iError,iError);
		return szErr;
	}
	return sError[i].pszErr;
}

SERROR sError[] = {
  {0,							"No error"},
  {WSABASEERR,         	"No error"},
  {WSAEINTR,        		"Interrupted system call"},
  {WSAEBADF,        		"Bad file number"},
  {WSAEACCES,       		"Permission denied"},
  {WSAEFAULT,       		"Bad address"},
//  {WSAEEXIST,		  		"File/Socket already exist"},
  {WSAEINVAL,       		"Invalid argument"},
  {WSAEMFILE,    	  		"Too many open files"},
  {WSAEWOULDBLOCK,     	"Operation would block"},
  {WSAEINPROGRESS,     	"Operation now in progress"},
  {WSAEALREADY,        	"Operation already in progress"},
  {WSAENOTSOCK,        	"Socket operation on non-socket"},
  {WSAEDESTADDRREQ,    	"Destination address required"},
  {WSAEMSGSIZE,        	"Message too long"},
  {WSAEPROTOTYPE,      	"Protocol wrong type for socket"},
  {WSAENOPROTOOPT,     	"Bad protocol option"},
  {WSAEPROTONOSUPPORT, 	"Protocol not supported"},
  {WSAESOCKTNOSUPPORT, 	"Socket type not supported"},
  {WSAEOPNOTSUPP,      	"Operation not supported on socket"},
  {WSAEPFNOSUPPORT,    	"Protocol family not supported"},
  {WSAEAFNOSUPPORT,    	"Address family not supported by protocol family"},
  {WSAEADDRINUSE,      	"Address already in use"},
  {WSAEADDRNOTAVAIL,   	"Can't assign requested address"},
  {WSAENETDOWN,        	"Network is down"},
  {WSAENETUNREACH,     	"Network is unreachable"},
  {WSAENETRESET,       	"Net dropped connection or reset"},
  {WSAECONNABORTED,    	"Software caused connection abort"},
  {WSAECONNRESET,      	"Connection reset by peer"},
  {WSAENOBUFS,         	"No buffer space available"},
  {WSAEISCONN,         	"Socket is already connected"},
  {WSAENOTCONN,        	"Socket is not connected"},
  {WSAESHUTDOWN,       	"Can't send after socket shutdown"},
  {WSAETOOMANYREFS,    	"Too many references, can't splice"},
  {WSAETIMEDOUT,       	"Connection timed out"},
  {WSAECONNREFUSED,    	"Connection refused"},
  {WSAELOOP,           	"Too many levels of symbolic links"},
  {WSAENAMETOOLONG,    	"File name too long"},
  {WSAEHOSTDOWN,       	"Host is down"},
  {WSAEHOSTUNREACH,    	"No Route to Host"},
  {WSAENOTEMPTY,       	"Directory not empty"},
  {WSAEPROCLIM,        	"Too many processes"},
  {WSAEUSERS,          	"Too many users"},
  {WSAEDQUOT,          	"Disc Quota Exceeded"},
  {WSAESTALE,          	"Stale NFS file handle"},
  {WSAEREMOTE,         	"Too many levels of remote in path"},
  {WSASYSNOTREADY,     	"Network SubSystem is unavailable"},
  {WSAVERNOTSUPPORTED, 	"WINSOCK DLL Version out of range"},
  {WSANOTINITIALISED,  	"Successful WSASTARTUP not yet performed"},
  {WSAHOST_NOT_FOUND,  	"Host not found"},
  {WSATRY_AGAIN,       	"Non-Authoritative Host not found"},
  {WSANO_RECOVERY,     	"Non-Recoverable errors: FORMERR, REFUSED, NOTIMP"},
  {WSANO_DATA,         	"Valid name, no data record of requested type"},
  {0x7fff,				  	0}
};
char *IPAddressString(DWORD dwIPAddress)
{
	static char sz[20];

	sprintf(sz, "%u.%u.%u.%u",
			((BYTE *)&dwIPAddress)[0],((BYTE *)&dwIPAddress)[1],
			((BYTE *)&dwIPAddress)[2],((BYTE *)&dwIPAddress)[3]);
	return sz;
}

time_t dwTargetTime;

// Set the timer for a wait in seconds
void SetSecondsTimer(WORD wTimeOut)
{
	dwTargetTime = time(0) + wTimeOut;
}

// Wait for keyboard or timeout
int DiagWait(void)
{
	int iKey;

	PollServer();
	if (kbhit())
	{
		iKey = Getch();
		if (iKey == 0x1b)	// Escape
		{
			Fail("&- user abort");
			return 1;
		}
		printf("Waiting for %s\n", szDiagnostic);
	}
	if (dwTargetTime && (dwTargetTime < time(0)))
	{
		Fail("&- timeout");
		return 1;
	}
	return 0;
}

// Start a test
// Save test name for diagnostics

int Start(char *szFormat, ...)
{
	va_list ap;

	if (bDiagnostic)
	{
		printf("Error in diagnostic: Start() without End() in: \n%s\n",
																				szDiagnostic);
		Exit(3);
	}
	iFail = 0;
	if (iControlBreak)
	{
		printf("User abort\n");
		Exit(1);
	}
	iNumberTests++;
	va_start(ap, szFormat);
	vsprintf(szDiagnostic, szFormat, ap);
	va_end(ap);
	if (iVerbose & 1)
		printf("%s ",szDiagnostic);
	bDiagnostic = 1;
	return 0;
}

// fail a test
int Fail (char *szFormat, ...)
{
	va_list ap;
	char szLine[400];

	if (!bDiagnostic)
	{
		printf("Error in diagnostic: Fail() without Start() in:\n%s\n",
																			szDiagnostic);
		Exit(3);
	}
	if (iFail)
		return 0;	// fail once only
	++iFailures;
	iFail = 1;
	if (!(iVerbose & 1))
		printf("%s ",szDiagnostic);
	if (szFormat)
	{
		va_start(ap, szFormat);
		vsprintf(szLine, szFormat + (*szFormat == '&' ? 1 : 0), ap);
		va_end(ap);
		if (*szFormat == '&')
			printf("%s - failed %s\n", szLine, GetErrorString(WSAGetLastError()));
		else
			printf("- failed %s\n%s\n", GetErrorString(WSAGetLastError()),szLine);
	}
	else
		printf("- failed %s\n", GetErrorString(WSAGetLastError()));
	if (iFailAction)
	{
		if (iFailAction < 2)
		{
			dwTargetTime = 0;		// don't time out
			printf("Press Esc to continue\n");
			while (!DiagWait());
			iFail = 0;
		}
		else {
			Exit(1);
		}
	}
	return 0;
}

// End a test. Print message if enabled
void End(void)
{
	if (!bDiagnostic)
	{
		printf("Error in diagnostic: End() without Start() in:\n%s\n",
																			szDiagnostic);
		Exit(3);
	}
#ifdef NULL_POINTER_TEST
	if (memcmp(arcNullData,(char *)0,sizeof(arcNullData)) != 0)
	{
		printf("Null pointer in %s\n",szDiagnostic);
		dwTargetTime = 0;		// don't time out
		printf("Press Esc to continue\n");
		while (!DiagWait());
		iFail = 0;
	}
#endif
	if ((iVerbose & 1) && !iFail)
	{
		printf("- Passed\n");
	}
	bDiagnostic = 0;
}

// server failed
void ServerFail(char *szFormat, ...)
{
	va_list ap;
	char szLine[400];

	++iServerFailures;
	if (bDebugServer)
	{
		va_start(ap, szFormat);
		vsprintf(szLine, szFormat + (*szFormat == '&' ? 1 : 0), ap);
		va_end(ap);
		printf("Server failed %s %s\n", szLine, GetErrorString(WSAGetLastError()));
	}
}


long GetMilliSeconds(void)
{
	return GET_MILLIS;
}

// Delay for iTicks milli-seconds
void Delay(int iMillis)
{
	long lTarget = (GetMilliSeconds() + iMillis + 1);

	while ((lTarget - GetMilliSeconds()) > 0) {
		PollServer();
	}
}

/*
		FD_SET(soListen,&fdsRead);
		FD_SET(soListen,&fdsWrite);
		FD_SET(soListen,&fdsExcept);
		lVal = select(0,&fdsRead,&fdsWrite,&fdsExcept,&sTimeVal);
		printf("listen select2 %ld r %d w %d e %d\n",lVal,FD_ISSET(soListen,&fdsRead),
						FD_ISSET(soListen,&fdsWrite),FD_ISSET(soListen,&fdsExcept));
*/
