//
//   $Id: XPING.C 1.4 2003/01/08 18:34:18 johnb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*
 * XPING.C
 *
 * Copyright (c) 2001 Datalight, Inc.
 * Datalight Confidential and Proprietary
 * All Rights Reserved
 * Portions Copyright GPvNO (c) 2001
 *
 * external pinger for Sockets
 *
 * REVISION HISTORY:
 * 05/21/2001 jmb made MSVC compatible
 * -----
 * Updated @ Date 27 October 1993, 17:52 by Gerrit van Niekerk
 * 1999-06-03 Solid code guidelines and documentation
 * 1999-10-98 Fixed bad report of first ping time
 * 2000-09-09 Fix divide overflow (unless you run it for 7 years!), search for API interrupt
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <dos.h>
#include <conio.h>
#include "compiler.h"
#include "api.h"

// prototypes
int SendPing(void);
int RecvPing(unsigned *puDelay);
unsigned GetClockTicks(void);

unsigned uAddrHigh;			//high word of host we are pinging
unsigned uAddrLow;			//low ...

unsigned long ulSentCount;	//number of Ping messages sent
unsigned long ulRecvCount; //number of Ping messages received

union REGS unRegs;			//used for API calls
unsigned uApiInt;				//sockets API interrupt number
char rgcInetID[] = "Inet"; //id to recognize sockets interrupt

main(int argc, char *argv[])
{
	void (far __interrupt *lpIntApiV)();
	char far *pId;
	int iChange; //whether change has occured and what kind of change
					 //0 - no change, 1 - received ping, 2 - sent ping, 3 - both sent and received
	int iCnt;

	//time fields (all in ticks)
	unsigned uTriggerClock;			//when again to send ping message
	unsigned uCurClock;     		//the current time
	unsigned uTickInterval = 10;	//interval between messages
	unsigned uDelayTime = 0;		//delay time (round-trip-time) of last message received
	unsigned uMaxDelayTime = 0; 	//maximum delay time
	unsigned uMinDelayTime = 0xffff;		//minimum delay time
	unsigned long ulTotalDelayTime = 0;	//total delay time

	//String with summery of ping response
	static char rgcResponse[80] = "no response";

	//buffers for string reps of messages (are combined in rgcResponse)
	char rgcLowString[20], rgcAveString[20], rgcHighString[20];

	// get sockets API interrupt number
	for (uApiInt = 0x20; uApiInt < 0x100; ++uApiInt)
	{
		lpIntApiV = _dos_getvect(uApiInt);
		if (lpIntApiV == 0)
			continue;
		pId = (char far *)lpIntApiV - 4;
		for (iCnt = 0; iCnt < 4; ++iCnt)
		{
			if (pId[iCnt] != rgcInetID[iCnt])
				break;
		}
		if (iCnt == 4)
			break;
	}
	if (uApiInt == 0x100)
	{
		printf("Sockets API not loaded\n");
		exit(1);
	}
	if (argc > 1)
	{
		//resolve host name
		unRegs.x.bx = FP_SEG(argv[1]);
		unRegs.x.cx = FP_OFF(argv[1]);
		unRegs.h.ah = API_GETIPADR;
		int86(uApiInt, &unRegs, &unRegs);
		if (unRegs.x.cx == 0 && unRegs.x.dx == 0)
		{
			printf("Unknown host: %s\n", argv[1]);
			exit(1);
		}
		uAddrHigh = unRegs.x.cx;
		uAddrLow = unRegs.x.dx;
	}
	else
	{
		//No host specified
		printf("Please specify host [and optional interval period]\n");
		exit(1);
	}
	if (argc > 2) //user specified intervals (for sending messages)
		uTickInterval = atoi(argv[2]);
	if (uTickInterval == 0)
		uTickInterval = 10;
	printf("Pinging %u.%u.%u.%u (Press any key to abort)\n",
					uAddrHigh >> 8, uAddrHigh & 0xff,
					uAddrLow >> 8, uAddrLow & 0xff);
	uTriggerClock = GetClockTicks() + 1;
	do
	{
		uCurClock = GetClockTicks();
		if ((int)(uTriggerClock - uCurClock) < 0)
		{
			//time to send a ping
			SendPing();
			uTriggerClock = GetClockTicks() + uTickInterval;
			iChange = 2;
		}
		iChange |= RecvPing(&uDelayTime);
		if (iChange)
		{
			//either sent or received a ping message (or both)
			if (iChange & 1)
			{
				//received ping message, recalculate statistics
				if (uDelayTime > uMaxDelayTime)
					uMaxDelayTime = uDelayTime;
				if (uDelayTime < uMinDelayTime)
					uMinDelayTime = uDelayTime;
				ulTotalDelayTime += uDelayTime;
				if (uMinDelayTime)
					sprintf(rgcLowString, "%lu", uMinDelayTime * 55l);
				else
					strcpy(rgcLowString, "<55");
				if (ulTotalDelayTime)
					sprintf(rgcAveString, "%lu", ulTotalDelayTime * 55l / ulRecvCount);
				else
					strcpy(rgcAveString, "<55");
				if (uMaxDelayTime)
					sprintf(rgcHighString, "%lu", uMaxDelayTime * 55l);
				else
					strcpy(rgcHighString, "<55");
				sprintf(rgcResponse, "min %s ave %s max %s ms", rgcLowString, rgcAveString, rgcHighString);
			}
			printf("\rSent: %lu, received %lu - %s      ", ulSentCount, ulRecvCount, rgcResponse);
			iChange = 0;
		}
	} while (kbhit() == 0);

	//flush all pending pings
	while (RecvPing(&uDelayTime) != 0);
	printf("\n");
	getch();
	if (ulRecvCount == 0)
		printf("No response!\n");
	return 0;
}

/*
Send a ping message to host. Put the current clock ticks in the sequence
number so that we can calculate a round-tip-time (delay) when this ping is
responded upon.

Returns:
	Zero
*/
int SendPing()
{
	unRegs.x.cx = uAddrHigh;
	unRegs.x.dx = uAddrLow;
	unRegs.x.bx = GetClockTicks();
	unRegs.h.ah = API_SENDPING;
	int86(uApiInt, &unRegs, &unRegs);
	++ulSentCount;
	return 0;
}

/*
See if there is a ping to receive. If there is, the round-trip-time (delay) is
calculated (in clock ticks) and stored at puDelay.

Arguments:
	puDelay - A pointer to where the delay is stored

Returns:
	1 if we received ping, 0 otherwise
*/
int RecvPing(unsigned *puDelay)
{
	unRegs.x.ax = API_GETPING << 8;
	int86(uApiInt, &unRegs, &unRegs);
	if (unRegs.x.ax != 0 && unRegs.x.cx == uAddrHigh && unRegs.x.dx == uAddrLow)
	{
		*puDelay = GetClockTicks() - unRegs.x.bx;
		++ulRecvCount;
		return 1;
	}
	return 0;
}

/*
Get the clock ticks stored at memory 0x40:0x6c.

Returns:
	The clock ticks for this hour
*/
unsigned GetClockTicks(void)
{
#ifdef _MSC_VER
 unsigned far *fpTicks = MK_FP(0x40, 0x6c);
 return *fpTicks;
#else
 return peek(0x40, 0x6c);
#endif
}
